#!/usr/bin/env python3

import argparse

parser = argparse.ArgumentParser(
    description="Filter TRUST4 barcode report file for noise like diffused mRNAs. Output to stdout"
)
parser.add_argument(
    "-b", help="barcode_report file", required=True, dest="barcode_report"
)
parser.add_argument("-a", help="annotation file", dest="annot")
parser.add_argument(
    "--highAbund",
    help="The minimum abundance to be regarded as potential source of diffusion",
    default=50.0,
    dest="highAbund",
)
parser.add_argument(
    "--diffuseFrac",
    help="The maximum fraction of the diffusion source abundance to be regarded as noise",
    default=0.02,
    dest="diffuseFrac",
)

args = parser.parse_args()

barcodeReport = args.barcode_report
barcodeInfo = {}
highAbundCdr3Info = {}
highAbund = float(args.highAbund)
diffuseFrac = float(args.diffuseFrac)
assembly = {}

fp = open(barcodeReport)
for line in fp:
    if line[0] == "#":
        continue
    cols = line.rstrip().split()
    chain1Cols = cols[2].split(",")
    chain2Cols = cols[3].split(",")
    # AGAGTGGTCTATCCTA-1	abT	TRBV6-6*01,TRBD2*01,TRBJ2-5*01,*,TGTGCCAGTCTACTTGGGGGGACCCAGTACTTC,CASLLGGTQYF,3.61,AGAGTGGTCTATCCTA-1_48281,100.00,0	TRAV10*01,*,TRAJ34*01,*,TGTGTGGTGAGCGCCCGCACCGACAAGCTCATCTTT,CVVSARTDKLIF,1.00,AGAGTGGTCTATCCTA-1_65563,100.00,0	*	*
    if len(chain1Cols) > 1 and float(chain1Cols[6]) >= highAbund:
        cdr3 = chain1Cols[4]
        abund = float(chain1Cols[6])
        if cdr3 not in highAbundCdr3Info:
            highAbundCdr3Info[cdr3] = {}
        highAbundCdr3Info[cdr3][cols[0]] = [0, abund]
    if len(chain1Cols) > 1:
        assembly[chain1Cols[7]] = chain1Cols[4]
        chain1Cols[6] = float(chain1Cols[6])

    if len(chain2Cols) > 1 and float(chain2Cols[6]) >= highAbund:
        cdr3 = chain2Cols[4]
        abund = float(chain2Cols[6])
        if cdr3 not in highAbundCdr3Info:
            highAbundCdr3Info[cdr3] = {}
        highAbundCdr3Info[cdr3][cols[0]] = [1, abund]
    if len(chain2Cols) > 1:
        assembly[chain2Cols[7]] = chain2Cols[4]
        chain2Cols[6] = float(chain2Cols[6])
    barcodeInfo[cols[0]] = {}
    barcodeInfo[cols[0]]["chain1"] = chain1Cols[:]
    barcodeInfo[cols[0]]["chain2"] = chain2Cols[:]
fp.close()

if args.annot is not None:
    fp = open(args.annot)
    for line in fp:
        header = line.rstrip()
        seq = fp.readline().rstrip()
        assemblyId = header.split()[0][1:]
        if assemblyId in assembly:
            assembly[assemblyId] = seq
    fp.close()

fp = open(barcodeReport)
for line in fp:
    if line[0] == "#":
        print(line.rstrip())
        continue

    cols = line.rstrip().split()
    chain1Cols = cols[2].split(",")
    chain2Cols = cols[3].split(",")

    testAgainst = {}
    if (
        len(chain1Cols) > 1
        and float(chain1Cols[6]) < highAbund
        and chain1Cols[4] in highAbundCdr3Info
    ):
        cdr3 = chain1Cols[4]
        for bc in highAbundCdr3Info[cdr3].keys():
            if highAbundCdr3Info[cdr3][bc][0] == 0 and highAbundCdr3Info[cdr3][bc][
                1
            ] * diffuseFrac > float(chain1Cols[6]):
                testAgainst[bc] = 1
    if (
        len(chain2Cols) > 1
        and float(chain2Cols[6]) < highAbund
        and chain2Cols[4] in highAbundCdr3Info
    ):
        cdr3 = chain2Cols[4]
        for bc in highAbundCdr3Info[cdr3]:
            if highAbundCdr3Info[cdr3][bc][0] == 1 and highAbundCdr3Info[cdr3][bc][
                1
            ] * diffuseFrac > float(chain2Cols[6]):
                testAgainst[bc] = 1
    flagFilter = 0
    for bc in testAgainst:
        testChain1 = barcodeInfo[bc]["chain1"]
        testChain2 = barcodeInfo[bc]["chain2"]

        for i in [1, 2]:
            c1 = chain1Cols
            c2 = testChain1
            if i == 2:
                c1 = chain2Cols
                c2 = testChain2
            if len(c1) > 1 and len(c2) > 1:
                if (
                    c2[6] * diffuseFrac > float(c1[6])
                    and assembly[c1[7]] in assembly[c2[7]]
                ):
                    flagFilter |= i
            elif len(c2) > 1:
                flagFilter |= i
    if flagFilter != 3:
        print(line.rstrip())
fp.close()



import math
import numpy as np

import plotly
import plotly.express as px
import plotly.graph_objects as go

from celescope.tools import utils

PLOTLY_CONFIG = {
    "displayModeBar": True,
    "staticPlot": False,
    "showAxisDragHandles": False,
    "modeBarButtons": [["toImage", "resetScale2d"]],
    "scrollZoom": False,
    "displaylogo": False,
}

COLORS = px.colors.qualitative.Plotly + px.colors.qualitative.Alphabet

LAYOUT = {
    "height": 313,
    "width": 400,
    "margin": {
        "l": 45,
        "r": 35,
        "b": 30,
        "t": 30,
    },
}


class Plotly_plot:
    def __init__(self, df):
        self._df = df
        self._fig = None

    def plotly_plot(self):
        return plotly.offline.plot(
            self._fig, include_plotlyjs=False, output_type="div", config=PLOTLY_CONFIG
        )

    def get_plotly_div(self):
        return self.plotly_plot()


class Tsne_plot(Plotly_plot):
    def __init__(self, df_tsne, feature_name, discrete=True):
        super().__init__(df_tsne)

        self.feature_name = feature_name
        self.discrete = discrete
        title_feature_name = feature_name[0].upper() + feature_name[1:]
        self.title = f"t-SNE plot Colored by {title_feature_name}"

        self._layout = {}
        self._dot_size = 4
        self._df["size"] = self._dot_size
        self._df["barcode_index"] = list(range(1, len(self._df) + 1))
        self._str_coord1 = "tSNE_1"
        self._str_coord2 = "tSNE_2"
        self.axes_config = {
            "showgrid": True,
            "gridcolor": "#F5F5F5",
            "showline": False,
            "ticks": None,
            "zeroline": True,
            "zerolinecolor": "black",
            "zerolinewidth": 0.7,
        }

        self.scatter_config = {
            "data_frame": df_tsne,
            "title": self.title,
            "x": self._str_coord1,
            "y": self._str_coord2,
            "size_max": self._dot_size,
            "hover_data": {
                self._str_coord1: False,
                self._str_coord2: False,
                self.feature_name: True,
                "barcode_index": True,
                "size": False,
            },
            "size": "size",
            "opacity": 0.9,
            "color": self.feature_name,
            "color_discrete_sequence": COLORS,
            "color_continuous_scale": px.colors.sequential.Jet,
        }

    def set_color_scale(self, color_scale):
        self.scatter_config["color_continuous_scale"] = color_scale

    @utils.add_log
    def get_plotly_div(self):
        if self.discrete:
            self.discrete_tsne_plot()
        else:
            self.continuous_tsne_plot()
        self.update_fig()

        return self.plotly_plot()

    @utils.add_log
    def discrete_tsne_plot(self):
        sum_df = self._df.groupby([self.feature_name]).agg("count").iloc[:, 0]
        percent_df = sum_df.transform(lambda x: round(x / sum(x) * 100, 2))
        res_dict = dict()
        res_list = []
        for cluster in sorted(self._df[self.feature_name].unique()):
            name = f"{cluster}({percent_df[cluster]}%)"
            res_dict[cluster] = name
            res_list.append(name)

        self._df[self.feature_name] = self._df[self.feature_name].map(res_dict)

        self._fig = px.scatter(
            **self.scatter_config, category_orders={self.feature_name: res_list}
        )

    @utils.add_log
    def continuous_tsne_plot(self):
        self._fig = px.scatter(
            **self.scatter_config,
        )

    def update_fig(self):
        self._fig.update_xaxes(title_text=self._str_coord1, **self.axes_config)

        self._fig.update_yaxes(title_text=self._str_coord2, **self.axes_config)

        self._fig.update_layout(
            self._layout,
            title={"text": self.title, "x": 0.5, "y": 0.95, "font": {"size": 15}},
            plot_bgcolor="#FFFFFF",
            hovermode="closest",
        )


class Tsne_dropdown_plot(Plotly_plot):
    def __init__(self, df_tsne, name, feature_name_list):
        super().__init__(df_tsne)

        self.name = name
        self.feature_name_list = feature_name_list
        self.title = f"t-SNE plot Colored by {self.name}"
        self.x_pos = 0.1
        self._layout = {}

        self._buttons = []
        self.x_label, self.y_label = "tSNE_1", "tSNE_2"

        self.x = df_tsne[self.x_label]
        self.y = df_tsne[self.y_label]
        self.axes_config = {
            "showgrid": True,
            "gridcolor": "#F5F5F5",
            "showline": False,
            "ticks": None,
            "zeroline": True,
            "zerolinecolor": "black",
            "zerolinewidth": 0.7,
        }

        x_ = math.ceil(max(abs(self.x)))
        y_ = math.ceil(max(abs(self.y)))
        self.x_range = [-x_, x_]
        self.y_range = [-y_, y_]

    def get_plotly_div(self):
        self.continuous_tsne_plot()
        self.update_fig()
        return self.plotly_plot()

    def continuous_tsne_plot(self):
        self._fig = go.Figure()

        # remove the cluster fraction "cluster 1(10%)"
        self._df["hover_text"] = "cluster " + self._df["cluster"].apply(
            lambda x: x.split("(")[0]
        )

        # Coordinates
        for i, feature in enumerate(self.feature_name_list):
            self._fig.add_trace(
                go.Scattergl(
                    x=self.x,
                    y=self.y,
                    mode="markers",
                    marker=dict(
                        size=4,
                        color=self._df[feature],
                        colorbar=dict(title="UMI"),
                        colorscale=px.colors.sequential.Jet,
                        opacity=0.9,
                    ),
                    name=feature,
                    hovertext=self._df["hover_text"],  # Text to display when hovering
                    hoverinfo="text",  # Display only the hover text on hover
                    visible=True if i == 0 else False,
                )
            )

    def update_fig(self):
        self._fig.update_xaxes(
            range=self.x_range, title_text=self.x_label, **self.axes_config
        )

        self._fig.update_yaxes(
            range=self.y_range, title_text=self.y_label, **self.axes_config
        )

        # Create dropdown buttons to toggle traces
        buttons = []
        for i, feature in enumerate(self.feature_name_list):
            visibility = [False] * len(self.feature_name_list)
            visibility[i] = True
            buttons.append(
                dict(
                    label=feature,
                    method="update",
                    args=[
                        {"visible": visibility},
                    ],
                )
            )

        self._fig.update_layout(
            self._layout,
            updatemenus=[
                {
                    "buttons": buttons,
                    "direction": "down",
                    "pad": {"r": 8, "t": 8},
                    "showactive": True,
                    "x": 1.1,
                    "xanchor": "right",
                    "y": 1.2,
                    "yanchor": "top",
                }
            ],
            annotations=[
                {
                    "text": "TAG type:",
                    "showarrow": False,
                    "x": self.x_pos,
                    "xref": "paper",
                    "xanchor": "right",
                    "y": 1.15,
                    "yref": "paper",
                    "font": {"size": 11},
                    "align": "left",
                }
            ],
            title={"text": self.title, "x": 0.5, "y": 0.95, "font": {"size": 15}},
            plot_bgcolor="#FFFFFF",
            hovermode="closest",
        )


class Pie_plot(Plotly_plot):
    def __init__(self, df_region):
        super().__init__(df_region)
        self.set_fig()

    def set_fig(self):
        layout = {
            "height": 300,
            "width": 400,
            "margin": {
                "l": 50,
                "r": 35,
                "b": 10,
                "t": 10,
            },
        }
        self._fig = px.pie(
            self._df,
            names="regions",
            values="values",
        )
        self._fig.update_traces(textposition="none")
        self._fig.update_layout(layout)


class Line_plot(Plotly_plot):
    def __init__(
        self,
        df_line,
        title=None,
        x_title=None,
        y_title=None,
        y_range=None,
        section=True,
    ):
        super().__init__(df_line)
        self.df_line = df_line
        self.title = title
        self.x_title = x_title
        self.y_title = y_title
        self.y_range = y_range
        self.section = section

        self.xaxes_config = {
            "showgrid": True,
            "gridcolor": "#F5F5F5",
            "linecolor": "black",
            "showline": True,
            "ticks": None,
            "tickmode": "linear",
            "tick0": 0,
            "dtick": 0.5,
        }

        if self.section:
            self.yaxes_config = {
                "showgrid": True,
                "gridcolor": "#F5F5F5",
                "linecolor": "black",
                "showline": True,
                "ticks": None,
                "rangemode": "tozero",
            }
        else:
            self.yaxes_config = {
                "showgrid": True,
                "gridcolor": "#F5F5F5",
                "linecolor": "black",
                "showline": True,
                "ticks": None,
                "range": self.y_range,
            }

        self.line_config = {
            "data_frame": df_line,
            "title": self.title,
            "x": self.x_title,
            "y": self.y_title,
        }

        self.line_plot()
        self.update_fig()

    @utils.add_log
    def line_plot(self):
        self._fig = px.line(
            **self.line_config,
        )

    def update_fig(self):
        self._fig.update_xaxes(**self.xaxes_config)

        self._fig.update_yaxes(**self.yaxes_config)

        self._fig.update_layout(
            LAYOUT,
            title={"x": 0.5, "y": 0.95, "font": {"size": 15}},
            yaxis_zeroline=True,
            showlegend=False,
            plot_bgcolor="#FFFFFF",
            hovermode="closest",
        )


class Bar_plot(Plotly_plot):
    def __init__(self, df_bar):
        super().__init__(df_bar)
        self.set_fig()

    def set_fig(self):
        self._fig = px.bar(
            x=[str(i) for i in list(self._df.head(10).ClonotypeID)],
            y=self._df.head(10).proportion.tolist(),
            labels={"x": "Clonotype ID", "y": "Fraction of Cells"},
            width=700,
            height=500,
        )
        self._fig.update_traces(
            marker_color="rgb(158,202,225)",
            marker_line_color="rgb(8,48,107)",
            marker_line_width=1.5,
            opacity=0.6,
        )
        self._fig.update_layout(
            title_text="Top 10 Clonotype Frequencies",
            title={"x": 0.5, "y": 0.9, "font": {"size": 20, "family": "San Serif"}},
            plot_bgcolor="#FFFFFF",
        )


class Conversion_plot(Plotly_plot):
    def __init__(self, df_bar):
        super().__init__(df_bar)
        np.random.seed(1)
        self.values, bins = np.histogram(df_bar["cell_pct"], bins=200)
        self.bin_centers = (bins[:-1] + bins[1:]) / 2
        self.set_fig()

    def set_fig(self):
        self._fig = px.bar(x=self.bin_centers, y=self.values)
        self._fig.update_layout(yaxis_type="log")
        self._fig.update_xaxes(title="cell percent")
        self._fig.update_yaxes(title="count")
        self._fig.update_layout(plot_bgcolor="rgba(0,0,0,0)")


class Substitution_plot(Plotly_plot):
    def __init__(self, df_bar):
        super().__init__(df_bar)
        self.set_fig()

    def set_fig(self):
        num4sample = 0
        colors4sample = {}
        num4x = 0

        self._fig = go.Figure()

        for sample in self._df["sample"].unique():
            legend_show = True
            colors4sample[sample] = COLORS[num4sample]
            num4sample += 1
            flag_x = "x" + str(num4x + 1)
            df_plot = self._df[self._df["sample"] == sample]
            num4x += 1

            self._fig.add_trace(
                go.Bar(
                    name=sample + "+",
                    x=df_plot["sample"],
                    y=df_plot["+"],
                    legendgroup=sample,
                    marker_color=colors4sample[sample],
                    marker_line_color="#FFFFFF",
                    showlegend=legend_show,
                    xaxis=flag_x,
                )
            )
            self._fig.add_trace(
                go.Bar(
                    name=sample + "-",
                    x=df_plot["sample"],
                    y=df_plot["-"],
                    legendgroup=sample,
                    showlegend=legend_show,
                    marker_color=colors4sample[sample],
                    marker_line_color="#FFFFFF",
                    opacity=0.3,
                    xaxis=flag_x,
                )
            )

        self._fig.update_layout(barmode="stack")

        per = 1 / (num4x + 1)
        gap4bar = per / len(self._df["sample"].unique())
        num4x = 0
        for _ in self._df["sample"].unique():
            if num4x == 0:
                flag_x = "xaxis"
            else:
                flag_x = "xaxis" + str(num4x + 1)
            anchor_x = "x" + str(num4x + 1)
            num4x += 1
            self._fig["layout"][flag_x] = dict(
                domain=[per * num4x, per * (num4x + 1) - gap4bar], anchor=anchor_x
            )

        self._fig.update_layout(plot_bgcolor="#FFFFFF")
        self._fig.update_xaxes(
            showgrid=False,
            linecolor="black",
            showline=True,
            ticks="outside",
            showticklabels=True,
        )
        self._fig.update_yaxes(
            showgrid=False, linecolor="black", showline=True, ticks="outside"
        )
        # width_num = 400 * (len(self._df['sample'].unique()) * len(self._df['sample'].unique())) / (5*12)  # set bar width
        self._fig.update_layout(height=500)  # , width=width_num)
        self._fig.update_layout(legend=dict(orientation="h"))
        self._fig.update_layout(
            legend=dict(
                yanchor="top",
                y=1.3,
                xanchor="left",
                x=0.07,
                valign="top",
            )
        )
        self._fig.update_layout(yaxis_title="Rates of nucleotide substitution (%)")


class Violin_plot(Plotly_plot):
    def __init__(self, df_vln, feature_name, color="#1f77b4", label=True):
        super().__init__(df_vln)
        self.feature_name = feature_name
        self.color = color
        self.label = label

        if self.label:
            self.y = "Fraction of labeled UMI"
        else:
            self.y = "Fraction of new RNA"
        self.set_fig()

    def set_fig(self):
        self._fig = go.Figure()
        self._fig.add_trace(
            go.Violin(
                y=self._df,
                box_visible=True,
                meanline_visible=True,
                line_color="black",
                fillcolor=self.color,
                opacity=0.6,
                x0=self.feature_name,
            )
        )
        self._fig.update_layout(yaxis_zeroline=True, showlegend=False)
        self._fig.update_layout(plot_bgcolor="#FFFFFF")
        self._fig.update_xaxes(
            showgrid=False, linecolor="black", showline=True, ticks=None
        )
        self._fig.update_yaxes(
            showgrid=False,
            linecolor="black",
            showline=True,
            ticks="outside",
            title_text=self.y,
            rangemode="tozero",
        )



import subprocess

import pysam
from celescope.tools.step import Step, s_common


class PathSeq(Step):
    def __init__(self, args, display_title=None):
        super().__init__(args, display_title)
        self.unmap_bam = f"{self.out_prefix}_unmapped.bam"
        self.downsample_bam = f"{self.out_prefix}_downsample.bam"
        self.addRG_bam = f"{self.out_prefix}_addRG.bam"
        self.pathseq_bam = f"{self.out_prefix}_pathseq.bam"
        self.pathseq_score_txt = f"{self.out_prefix}_pathseq_score.txt"
        self.filter_metrics_txt = f"{self.out_prefix}_filter_metrics.txt"

    def get_unmap_bam(self):
        cmd = (
            f"samtools view -b -h -f 4 "
            f"{self.args.input_bam} "
            f"> {self.unmap_bam} "
        )
        subprocess.check_call(cmd, shell=True)

    @staticmethod
    def count_reads(bam_file):
        count = 0
        with pysam.AlignmentFile(bam_file, "rb") as bam:
            for read in bam:
                count += 1
        return count

    def downsample(self):
        read_count = self.count_reads(self.unmap_bam)
        downsample_rate = self.args.downsample_reads / read_count
        downsample_rate = min(downsample_rate, 1.0)
        cmd = (
            f"samtools view -b -s {downsample_rate} "
            f"{self.unmap_bam} "
            f"> {self.downsample_bam} "
        )
        subprocess.check_call(cmd, shell=True)

    def picard_addRG(self):
        cmd = (
            f"picard AddOrReplaceReadGroups "
            f"INPUT={self.downsample_bam} "
            f"OUTPUT={self.addRG_bam} "
            f"RGID={self.sample}_rgid "
            f"RGLB={self.sample}_rglb "
            f"RGPL=illumina "
            f"RGPU={self.sample}_rgpu "
            f"RGSM={self.sample}_rgsm "
        )
        subprocess.check_call(cmd, shell=True)

    def run_pathseq(self):
        cmd = (
            f"gatk PathSeqPipelineSpark "
            f"--input {self.addRG_bam} "
            f"--filter-bwa-image {self.args.filter_bwa_image} "
            f"--kmer-file {self.args.kmer_file} "
            f"--microbe-bwa-image {self.args.microbe_bwa_image} "
            f"--microbe-dict {self.args.microbe_dict} "
            f"--taxonomy-file {self.args.microbe_taxonomy_file} "
            f"--java-options -Xmx100g "
            f"--min-score-identity {self.args.min_score_identity} "
            f"--min-clipped-read-length {self.args.min_clipped_read_length} "
            f"--is-host-aligned true "
            f"--filter-duplicates false "
            f"--output {self.pathseq_bam} "
            f"--scores-output {self.pathseq_score_txt} "
            f"--filter-metrics {self.filter_metrics_txt} "
        )
        subprocess.check_call(cmd, shell=True)

    def run(self):
        self.get_unmap_bam()
        self.downsample()
        self.picard_addRG()
        self.run_pathseq()


def pathseq(args):
    with PathSeq(args) as runner:
        runner.run()


def get_opts_pathseq(parser, sub_program):
    parser.add_argument("--filter_bwa_image", help="filter_bwa_image", required=True)
    parser.add_argument("--kmer_file", help="kmer_file", required=True)
    parser.add_argument("--microbe_bwa_image", help="microbe_bwa_image", required=True)
    parser.add_argument("--microbe_dict", help="microbe_dict", required=True)
    parser.add_argument(
        "--microbe_taxonomy_file", help="microbe_taxonomy_file", required=True
    )
    parser.add_argument("--min_score_identity", help="min_score_identity", default=0.7)
    parser.add_argument(
        "--min_clipped_read_length", help="min_clipped_read_length", default=60
    )
    parser.add_argument(
        "--downsample_reads",
        help="It is strongly recommended that there are <10M total microbial reads.",
        default=10**7,
        type=int,
    )
    if sub_program:
        parser.add_argument("--input_bam", help="input bam file", required=True)
        parser = s_common(parser)
    return parser



import scanpy as sc
import numpy as np
import pandas as pd

from celescope.tools import utils
from celescope.__init__ import HELP_DICT
from celescope.rna.mkref import Mkref_rna
from celescope.tools.step import Step, s_common

# markers adjust p_value
PVAL_CUTOFF = 0.05
# scanpy mitochonrial variable name
MITO_VAR = "mito"
NORMALIZED_LAYER = "normalised"
RESOLUTION = 1.2
N_PCS = 25
MITO_GENE_PERCENT_LIST = [5, 10, 15, 20, 50]
# output marker top n in html
MARKER_TOP_N = 100


def read_tsne(tsne_file):
    df = pd.read_csv(tsne_file, sep="\t", index_col=0)
    df.index.names = ["barcode"]
    return df


def format_df_marker(df_marker):
    avg_logfc_col = "avg_log2FC"  # seurat 4
    if "avg_logFC" in df_marker.columns:  # seurat 2.3.4
        avg_logfc_col = "avg_logFC"
    df_marker = df_marker.loc[
        :, ["cluster", "gene", avg_logfc_col, "pct.1", "pct.2", "p_val_adj"]
    ]
    df_marker["cluster"] = df_marker["cluster"].apply(lambda x: f"cluster {x}")
    df_marker = df_marker[df_marker["p_val_adj"] < PVAL_CUTOFF]

    return df_marker


def get_opts_analysis(parser, sub_program):
    parser.add_argument("--genomeDir", help=HELP_DICT["genomeDir"], required=True)
    if sub_program:
        parser.add_argument(
            "--matrix_file",
            help="Required. Matrix directory path.",
            required=True,
        )
        parser = s_common(parser)


class Scanpy_wrapper(Step):
    def __init__(self, args, display_title=None):
        super().__init__(args, display_title=display_title)

        # data
        self.adata = sc.read_10x_mtx(
            args.matrix_file,
            var_names="gene_symbols",
        )
        self.adata.layers["raw"] = self.adata.X.copy()
        self.mt_gene_list = Mkref_rna.get_config(args.genomeDir)["files"][
            "mt_gene_list"
        ]

        # out
        self.df_marker_file = f"{self.outdir}/markers.tsv"
        self.df_marker_raw_file = f"{self.outdir}/markers_raw.tsv"
        self.df_tsne_file = f"{self.outdir}/tsne_coord.tsv"
        self.h5ad_file = f"{self.outdir}/rna.h5ad"
        self.outs = [self.df_marker_file, self.df_tsne_file, self.h5ad_file]

    @utils.add_log
    def calculate_qc_metrics(self):
        if self.mt_gene_list:
            mito_genes, _ = utils.read_one_col(self.mt_gene_list)
            self.adata.var[MITO_VAR] = self.adata.var_names.map(
                lambda x: True if x in mito_genes else False
            )
            # if not astype(bool), it will be type object and raise an error
            # https://github.com/theislab/anndata/issues/504
            self.adata.var[MITO_VAR] = self.adata.var[MITO_VAR].astype(bool)
        else:
            self.adata.var[MITO_VAR] = self.adata.var_names.str.upper().str.startswith(
                "MT-"
            )

        sc.pp.calculate_qc_metrics(
            self.adata,
            qc_vars=[MITO_VAR],
            percent_top=None,
            use_raw=False,
            log1p=False,
            inplace=True,
        )

    @utils.add_log
    def write_mito_stats(self):
        mt_pct_var = f"pct_counts_{MITO_VAR}"
        total_cell_number = self.adata.n_obs

        for mito_gene_percent in MITO_GENE_PERCENT_LIST:
            cell_number = sum(self.adata.obs[mt_pct_var] > mito_gene_percent)
            fraction = round(cell_number / total_cell_number * 100, 2)
            self.add_metric(
                name=f"Fraction of cells have mito gene percent>{mito_gene_percent}%",
                value=f"{fraction}%",
            )

    @utils.add_log
    def normalize(self):
        """
        sc.pp.normalize_per_cell() and sc.pp.log1p()
        """

        sc.pp.normalize_total(
            self.adata,
            target_sum=1e4,
            inplace=True,
        )
        sc.pp.log1p(
            self.adata,
        )
        self.adata.layers[NORMALIZED_LAYER] = self.adata.X

    @utils.add_log
    def hvg(self):
        """
        Wrapper function for sc.highly_variable_genes()
        """
        sc.pp.highly_variable_genes(
            self.adata,
            layer=None,
            n_top_genes=None,
            min_disp=0.5,
            max_disp=np.inf,
            min_mean=0.0125,
            max_mean=3,
            span=0.3,
            n_bins=20,
            flavor="seurat",
            subset=False,
            inplace=True,
            batch_key=None,
            check_values=True,
        )

    @utils.add_log
    def scale(self):
        """
        Wrapper function for sc.pp.scale
        """
        sc.pp.scale(
            self.adata,
            zero_center=True,
            max_value=10,
            copy=False,
            layer=None,
            obsm=None,
        )

    @utils.add_log
    def pca(self):
        """
        Wrapper function for sc.pp.pca
        """
        sc.pp.pca(
            self.adata,
            n_comps=50,
            zero_center=True,
            svd_solver="auto",
            random_state=0,
            return_info=False,
            use_highly_variable=True,
            dtype="float32",
            copy=False,
            chunked=False,
            chunk_size=None,
        )

    @utils.add_log
    def neighbors(
        self,
    ):
        """
        Wrapper function for sc.pp.neighbors(), for supporting multiple n_neighbors
        """
        sc.pp.neighbors(
            self.adata,
            n_neighbors=15,
            n_pcs=N_PCS,
            use_rep=None,
            knn=True,
            random_state=0,
            method="umap",
            metric="euclidean",
            key_added=None,
            copy=False,
        )

    @utils.add_log
    def tsne(self):
        """
        Wrapper function for sc.tl.tsne, for supporting named slot of tsne embeddings
        """
        sc.tl.tsne(
            self.adata,
            n_pcs=N_PCS,
            copy=False,
        )

    @utils.add_log
    def umap(
        self,
    ):
        """
        Wrapper function for sc.tl.umap, for supporting named slot of umap embeddings
        """
        sc.tl.umap(
            self.adata,
            min_dist=0.5,
            spread=1.0,
            n_components=2,
            maxiter=None,
            alpha=1.0,
            gamma=1.0,
            negative_sample_rate=5,
            init_pos="spectral",
            random_state=0,
            a=None,
            b=None,
            copy=False,
            method="umap",
            neighbors_key=None,
        )

    @utils.add_log
    def leiden(self):
        """
        Wrapper function for sc.tl.leiden
        """
        sc.tl.leiden(
            self.adata,
            resolution=RESOLUTION,
            restrict_to=None,
            random_state=0,
            key_added="cluster",
            adjacency=None,
            directed=True,
            use_weights=True,
            n_iterations=-1,
            partition_type=None,
            neighbors_key=None,
            obsp=None,
            copy=False,
        )

    @utils.add_log
    def find_marker_genes(self):
        """
        Wrapper function for sc.tl.rank_genes_groups
        """
        sc.tl.rank_genes_groups(
            self.adata,
            "cluster",
            reference="rest",
            pts=True,
            method="wilcoxon",
            use_raw=False,
            layer=NORMALIZED_LAYER,
        )

    @utils.add_log
    def write_markers(self):
        """
        write only p_val_adj < PVAL_CUTOFF to avoid too many markers
        """
        df_markers = sc.get.rank_genes_groups_df(
            self.adata, group=None, pval_cutoff=PVAL_CUTOFF
        )
        df_markers = df_markers[df_markers["logfoldchanges"].notna()]
        markers_name_dict = {
            "group": "cluster",
            "names": "gene",
            "logfoldchanges": "avg_log2FC",
            "pvals": "p_val",
            "pvals_adj": "p_val_adj",
            "pct_nz_group": "pct.1",
            "pct_nz_reference": "pct.2",
        }
        df_markers = df_markers.rename(markers_name_dict, axis="columns")
        df_markers["cluster"] = df_markers["cluster"].map(lambda x: int(x) + 1)
        df_markers = df_markers.loc[df_markers["p_val_adj"] < PVAL_CUTOFF,]
        df_markers.to_csv(self.df_marker_raw_file, index=None, sep="\t")

        df_markers_filter = (
            df_markers.loc[df_markers["avg_log2FC"] > 0]
            .sort_values("p_val_adj")
            .groupby("cluster")
            .head(100)
        )
        df_markers_filter = df_markers_filter.round(
            {
                "avg_log2FC": 3,
                "pct.1": 3,
                "pct.2": 3,
            }
        )
        df_markers_filter.to_csv(self.df_marker_file, index=None, sep="\t")

    @utils.add_log
    def write_tsne(self):
        df_tsne = self.adata.obsm.to_df()[["X_tsne1", "X_tsne2"]]
        df_tsne["cluster"] = self.adata.obs.cluster
        df_tsne["Gene_Counts"] = self.adata.obs.n_genes_by_counts
        tsne_name_dict = {"X_tsne1": "tSNE_1", "X_tsne2": "tSNE_2"}
        df_tsne = df_tsne.rename(tsne_name_dict, axis="columns")
        df_tsne["cluster"] = df_tsne["cluster"].map(lambda x: int(x) + 1)
        df_tsne.to_csv(self.df_tsne_file, sep="\t")

    @utils.add_log
    def write_h5ad(self):
        self.adata.write(self.h5ad_file)

    @utils.add_log
    def run(self):
        self.calculate_qc_metrics()
        self.write_mito_stats()
        self.normalize()
        self.hvg()
        self.scale()
        self.pca()
        self.neighbors()
        self.tsne()
        self.umap()
        self.leiden()
        self.find_marker_genes()

        self.write_markers()
        self.write_tsne()
        self.write_h5ad()

    def get_df(self):
        """
        return df_tsne, df_marker
        """
        df_tsne = read_tsne(self.df_tsne_file)
        df_marker = pd.read_csv(self.df_marker_file, sep="\t")
        df_marker = format_df_marker(df_marker)
        return df_tsne, df_marker


def get_opts_analysis_match(parser, sub_program):
    """
    Do not perform analysis. Only read data from scRNA-seq match_dir.
    """
    if sub_program:
        parser.add_argument("--match_dir", help=HELP_DICT["match_dir"])
        parser.add_argument("--tsne_file", help=HELP_DICT["tsne_file"])
        parser.add_argument("--df_marker_file", help=HELP_DICT["df_marker_file"])

        parser = s_common(parser)


class Report_runner(Step):
    def __init__(self, args, display_title=None):
        super().__init__(args, display_title=display_title)

    def add_marker_help(self):
        self.add_help_content(
            name="Marker Genes by Cluster",
            content="differential expression analysis based on the non-parameteric Wilcoxon rank sum test",
        )
        self.add_help_content(
            name="avg_log2FC",
            content="log fold-change of the average expression between the cluster and the rest of the sample",
        )
        self.add_help_content(
            name="pct.1",
            content="The percentage of cells where the gene is detected in the cluster",
        )
        self.add_help_content(
            name="pct.2",
            content="The percentage of cells where the gene is detected in the rest of the sample",
        )
        self.add_help_content(
            name="p_val_adj",
            content="Adjusted p-value, based on bonferroni correction using all genes in the dataset",
        )

    @staticmethod
    def get_df_file(match_dir):
        """
        return df_tsne_file, df_marker_file
        """
        match_dict = utils.parse_match_dir(match_dir)
        df_tsne_file = match_dict["tsne_coord"]
        df_marker_file = match_dict.get("markers", None)
        return df_tsne_file, df_marker_file

    def get_df(self):
        """
        return df_tsne, df_marker
        """
        if utils.check_arg_not_none(self.args, "match_dir"):
            df_tsne_file, df_marker_file = self.get_df_file(self.args.match_dir)
        elif utils.check_arg_not_none(self.args, "tsne_file"):
            df_tsne_file = self.args.tsne_file
            df_marker_file = self.args.df_marker_file
        else:
            raise ValueError("match_dir or tsne_file must be specified")
        df_tsne = read_tsne(df_tsne_file)
        if df_marker_file:
            df_marker = pd.read_csv(df_marker_file, sep="\t")
            df_marker = format_df_marker(df_marker)
        else:
            df_marker = None
        return df_tsne, df_marker

    def run(self):
        pass



import unittest
import os

import celescope.snp.variant_calling as vc
from tests.unittests.__init__ import TEST_DIR_ROOT


class Test_snp(unittest.TestCase):
    def setUp(self):
        os.chdir(f"{TEST_DIR_ROOT}/snp/")
        self.sample = "test1"
        self.vcf_file = f"{self.sample}/07.variant_calling/{self.sample}_merged.vcf"

    @unittest.skip("skip")
    def test_cell_UMI(self):
        cid = 7996
        outdir = f"{self.sample}/07.variant_calling/"
        df_UMI = vc.cell_UMI(cid, outdir, self.vcf_file)
        print(df_UMI)

    @unittest.skip("skip")
    def test_cell_UMI_2(self):
        cid = 2301
        os.chdir("/SGRNJ03/randd/RD20081701_SCOPEv2_Dynaseq/20210820/")
        sample = "Mito_9geneMix_0812"
        outdir = f"{sample}/07.variant_calling/"
        vcf_file = f"{sample}/07.variant_calling/{sample}_merged.vcf"
        df_UMI = vc.cell_UMI(cid, outdir, vcf_file)
        print(df_UMI)


if __name__ == "__main__":
    unittest.main()



import subprocess
import sys

from celescope.tools import utils
from celescope.tools.make_ref import MakeRef_STAR


def parse_attributes(attrs):
    dic = {}
    for attr_str in attrs.split(";"):
        if attr_str:
            attr, val = attr_str.split("=")
            val = set(val.split(","))
            dic[attr] = val
    return dic


class Mkref_rna(MakeRef_STAR):
    def __init__(self, genome_type, args):
        super().__init__(genome_type, args)
        self.files["gtf"] = self.args.gtf
        self.files["mt_gene_list"] = self.args.mt_gene_list

    @utils.add_log
    def build_rna_star_index(self):
        SA = self._get_SA()
        cmd = (
            f"STAR \\\n"
            f"--runMode genomeGenerate \\\n"
            f"--runThreadN {self.args.thread} \\\n"
            f"--genomeDir ./ \\\n"
            f"--genomeFastaFiles {self.args.fasta} \\\n"
            f"--sjdbGTFfile {self.args.gtf} \\\n"
            f"--sjdbOverhang 100 \\\n"
            f"--genomeSAindexNbases {SA} \\\n"
        )
        if self.STAR_param:
            cmd += " " + self.args.STAR_param
        sys.stderr.write(cmd + "\n")
        if not self.dry_run:
            subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def run(self):
        self.build_rna_star_index()


def mkref(args):
    genome_type = "rna"
    with Mkref_rna(genome_type, args) as runner:
        runner.run()


def get_opts_mkref(parser, sub_program):
    MakeRef_STAR.opts(parser, sub_program)
    if sub_program:
        parser.add_argument("--gtf", help="Required. Gtf file name.", required=True)
        parser.add_argument(
            "--mt_gene_list",
            help="""Mitochondria gene list file name. This file is a plain text file with one gene per line. 
If not provided, will use `MT-` and `mt-` to determine mitochondria genes.""",
            default="None",
        )
        parser.add_argument(
            "--attributes",
            help="Attributes to keep. Example: `gene_biotype=protein_coding,lncRNA,antisense;`",
            default="gene_biotype=protein_coding,lncRNA,antisense,IG_LV_gene,IG_V_gene,IG_V_pseudogene,IG_D_gene,IG_J_gene,IG_J_pseudogene,IG_C_gene,IG_C_pseudogene,TR_V_gene,TR_V_pseudogene,TR_D_gene,TR_J_gene,TR_J_pseudogene,TR_C_gene;",
        )



#!/usr/bin/env python3

import sys
import argparse
import math

isotypeRanks = {
    "IGHM": 0,
    "IGHD": 1,
    "IGHG3": 2,
    "IGHG1": 3,
    "IGHA1": 4,
    "IGHG2": 5,
    "IGHG4": 6,
    "IGHE": 7,
    "IGHA2": 8,
    "*": 9,
    ".": 9,
}
isotypeOrder = [
    "IGHM",
    "IGHD",
    "IGHG3",
    "IGHG1",
    "IGHA1",
    "IGHG2",
    "IGHG4",
    "IGHE",
    "IGHA2",
]
chainOrder = ["IGH", "IGK", "IGL", "TRA", "TRB", "TRG", "TRD"]


def GetChainType(v, j, c):
    s = ""
    if c != "*" and c != ".":
        s = c
    elif j != "*" and j != ".":
        s = j
    elif v != "*" and v != ".":
        s = v
    else:
        return -1

    if s[0:3] == "IGH":
        return (0, isotypeRanks[c])
    elif s[0:3] == "IGK":
        return 1
    elif s[0:3] == "IGL":
        return 2
    elif s[0:3] == "TRA":
        return 3
    elif s[0:3] == "TRB":
        return 4
    elif s[0:3] == "TRG":
        return 5
    elif s[0:3] == "TRD":
        return 6
    else:
        return -1


def ComputeRichness(rep):
    return len(rep)


def ComputeCPK(rep):
    if len(rep) == 0:
        return "NA"
    return len(rep) / sum(list(rep.values())) * 1000


def ComputeEntropy(rep):
    if len(rep) == 0:
        return "NA"
    total = sum(list(rep.values()))
    return sum([-x / total * math.log(x / total) for x in rep.values()])


def ComputeClonality(rep):
    if len(rep) <= 1:
        return "NA"
    return 1 - ComputeEntropy(rep) / math.log(len(rep))


def OutputChain(rep, name):
    outputList = [name]
    outputList.append(sum(list(rep.values())))
    outputList.append(ComputeRichness(rep))
    outputList.append(ComputeCPK(rep))
    outputList.append(ComputeEntropy(rep))
    outputList.append(ComputeClonality(rep))
    print("\t".join([str(x) for x in outputList]))


def ProcessImmuneRepertoire(immrep):
    print("\t".join(["#chain", "Abundance", "Richness", "CPK", "Entropy", "Clonality"]))
    # IGH overall
    tmp = {}
    for i in range(10):
        for c in immrep[(0, i)]:
            if c not in tmp:
                tmp[c] = 0
            tmp[c] += immrep[(0, i)][c]
    OutputChain(tmp, "IGH")
    # IGH chains
    for i in range(9):
        OutputChain(immrep[(0, i)], isotypeOrder[i])
    for i in range(1, 7):
        OutputChain(immrep[i], chainOrder[i])


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Immune repertoire diversity statistics"
    )
    parser.add_argument("-r", help="repertoire file", dest="repfile", required=True)
    parser.add_argument(
        "-f",
        help="repertoire file format (TRUST4_report, TRUST4_barcode_report,...)",
        dest="format",
        default="TRUST4_report",
    )
    parser.add_argument(
        "--ntaa",
        help="Use nucleotide(nt) or amino acids(aa)",
        dest="ntaa",
        default="aa",
    )
    args = parser.parse_args()

    immrep = {}
    for i in range(10):
        immrep[(0, i)] = {}
    for i in range(1, 7):
        immrep[i] = {}

    fp = open(args.repfile)
    if args.format == "TRUST4_report":
        for line in fp:
            if line[0] == "#" or line[0:5] == "count":
                continue
            cols = line.rstrip().split()
            chainType = GetChainType(cols[4], cols[6], cols[7])
            if (
                "_" in cols[3]
                or cols[3] == "partial"
                or "?" in cols[3]
                or chainType == -1
            ):
                continue
            if cols[3] not in immrep[chainType]:
                immrep[chainType][cols[3]] = 0
            immrep[chainType][cols[3]] += int(cols[0])
    elif args.format == "TRUST4_barcode_report":
        for line in fp:
            if line[0] == "#" or line[0:5] == "count":
                continue
            mainCols = line.rstrip().split()
            for i in [2, 3]:
                if mainCols[i] == "*":
                    continue
                cols = mainCols[i].split(",")
                chainType = GetChainType(cols[0], cols[2], cols[3])
                if (
                    "_" in cols[5]
                    or cols[5] == "partial"
                    or "?" in cols[5]
                    or chainType == -1
                ):
                    continue
                if cols[5] not in immrep[chainType]:
                    immrep[chainType][cols[5]] = 0
                immrep[chainType][cols[5]] += 1
    else:
        print("Unknown format " + args.format)
        sys.exit(1)
    fp.close()

    ProcessImmuneRepertoire(immrep)



STEPS = [
    "mkref",
    "sample",
    "barcode",
    "consensus",
    "mapping_vdj",
    "count_vdj",
]
__ASSAY__ = "vdj"

CHAINS = {
    "TCR": ["TRA", "TRB"],
    "BCR": ["IGH", "IGL", "IGK"],
}

PAIRS = {
    "TCR": [["TRA", "TRB"]],
    "BCR": [["IGH", "IGK"], ["IGH", "IGL"]],
}



import pandas as pd
import pysam
import copy
import subprocess
import json

from celescope.tools import analysis_wrapper
from collections import defaultdict
from celescope.tools import utils
from celescope.tools.capture.threshold import Auto
from celescope.tools.step import Step, s_common
from celescope.flv_trust4.__init__ import CHAIN, PAIRED_CHAIN, TOOLS_DIR
from celescope.tools.emptydrop_cr import get_plot_elements


ASSEMBLE_SUFFIX = "assembled_reads.fa"
ANNOTATE_SUFFIX = "annotate.fa"
TRUST_REPORT_SUFFIX = "filter_report.tsv"
BARCODE_REPORT_SUFFIX = "barcode_report.tsv"
BARCODE_REPORT_FILTER_SUFFIX = "barcode_filter_report.tsv"


def target_cell_calling(
    df_UMI_sum,
    expected_target_cell_num=3000,
    target_barcodes=None,
    weight=6,
    coef=5,
    percentile=85,
    umi_col="umis",
):
    """
    Args:
        df_UMI_sum: A dataframe with columns highest umi's contig and UMI.

    Returns:
        target_contigs_id: list
    >>> df_UMI_sum = pd.DataFrame({"contig_id": ["A", "B", "C", "D", "E"], "UMI": [1, 2, 1, 30, 40]})
    >>> target_contigs_id = target_cell_calling(df_UMI_sum, expected_target_cell_num=5, percentile=80, coef=5, target_barcodes=["A", "C"])
    >>> target_contigs_id == {'A_1', 'C_1', 'D_1', 'E_1'}
    True
    """
    if target_barcodes is not None:
        target_barcodes = {i for i in target_barcodes}
    umi_threshold = Auto(
        list(df_UMI_sum[umi_col]),
        expected_cell_num=expected_target_cell_num,
        coef=coef,
        percentile=percentile,
    ).run()

    # avoid change the original dataframe
    df_temp = df_UMI_sum.copy()
    if target_barcodes:
        df_temp[umi_col] = df_temp.apply(
            lambda row: row[umi_col] * weight
            if row["barcode"] in target_barcodes
            else row[umi_col],
            axis=1,
        )

    target_contigs = set(df_temp.loc[df_temp[umi_col] >= umi_threshold].contig_id)

    return target_contigs


class Summarize(Step):
    """
    ## Features
    - CDR3 filtering: contain stop condon, length <=5, etc..

    - If barcode A's two chains CDR3s are identical to another barcode B, and A's chain abundance is significantly lower than B's, filter A.

    - If `--target_cell_barcode` is provided, the UMI counts of all contigs originated from target cells are multiplied by a weight(default: 6.0) to better distinguish signal from background noise. `--target_cell_barcode` comes from the cell type annotation results of the RNA library.

    - Cell-calling is similar to the rna cell-calling algorithm.

    ## Output
    - `04.summarize/clonetypes.tsv` High-level description for each clonotype.

    - `04.summarize/{sample}_all_contig.csv` High-level and detailed annotation for each contig.

    - `04.summarize/{sample}_all_contig.fasta` All assembled contig sequences.

    - `04.summarize/{sample}_filtered_contig.csv` High-level annotations of each cellular contig after filter. This is a subset of {sample}_all_contig.csv.

    - `04.summarize/{sample}_filtered_contig.fasta` Assembled contig sequences after filter.
    """

    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)

        self.seqtype = args.seqtype
        self.species = args.species
        self.fq2 = args.fq2
        self.diffuseFrac = args.diffuseFrac
        self.assembled_fa = f"{args.assemble_out}/{self.sample}_{ASSEMBLE_SUFFIX}"
        self.trust_report = f"{args.assemble_out}/{self.sample}_{TRUST_REPORT_SUFFIX}"
        self.annot = f"{args.assemble_out}/{self.sample}_{ANNOTATE_SUFFIX}"

        self.match_dict = utils.parse_match_dir(args.match_dir)
        self.chains, self.paired_groups = self._parse_seqtype(self.seqtype)

        # if --diffuseFrac provided
        if self.diffuseFrac:
            self.barcode_report = (
                f"{args.assemble_out}/{self.sample}_{BARCODE_REPORT_FILTER_SUFFIX}"
            )
        else:
            self.barcode_report = (
                f"{args.assemble_out}/{self.sample}_{BARCODE_REPORT_SUFFIX}"
            )

        self.coef = int(args.coef)
        self.target_weight = args.target_weight

        """
        There are three method. 
        1. --expected_target_cell_num, Expected assembled T or B cell number. eg 3000
        2. --target_cell_barcode Auto, Based on auto assigned cell type annotation results of the RNA library.
        3. --target_cell_barcode Absolute path of a plain text file with one barcode per line.(Recommend, need manual-assign first)
        """
        if not args.target_cell_barcode:
            self.target_barcodes = None
            self.expected_target_cell_num = args.expected_target_cell_num
        elif args.target_cell_barcode == "Auto":
            self.target_barcodes, self.expected_target_cell_num = self._auto_assign()
        else:
            self.target_barcodes, self.expected_target_cell_num = utils.read_one_col(
                args.target_cell_barcode
            )

    @staticmethod
    def _parse_seqtype(seqtype):
        """Parse BCR or TCR

        :param seqtype
        :return CHAIN[seqtype]: 'TCR': ['TRA', 'TRB'], 'BCR': ['IGH', 'IGL', 'IGK']
                    PAIRED_CHAIN[seqtype]: ['TRA_TRB'], 'BCR': ['IGH_IGL', 'IGH_IGK']
        """
        return CHAIN[seqtype], PAIRED_CHAIN[seqtype]

    @staticmethod
    @utils.add_log
    def convert_barcode_report(barcode_report, outdir):
        cmd = (
            f"perl {TOOLS_DIR}/trust-barcoderep-to-10X.pl "
            f"{barcode_report} "
            f"{outdir} "
        )
        Summarize.convert_barcode_report.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    def add_cell_num_metric(self, df, name):
        """
        add cell number after each filtering step to .metrics.json
        do not show in HTML report
        """
        cell_num = len(set(df[df["productive"]].barcode))
        self.add_metric(name, cell_num, show=False)
        return cell_num

    @utils.add_log
    def _auto_assign(self):
        """
        --target_cell_barcode Auto
        Auto assign cell type of the RNA library based on marker gene file.
        target_cell_barcode requires >=2 identified marker genes of T/B cells in top10 marker genes for each cluster.
        Return target_barcodes, expected_target_cell_num.
        """

        with open(f"{TOOLS_DIR}/Immune_marker.json", "r") as f:
            marker_dict = json.load(f)

        report_runner = analysis_wrapper.Report_runner(self.args)
        df_tsne, df_marker = report_runner.get_df()

        bc_cluster_dict = (
            df_tsne.reset_index()
            .groupby("cluster")["barcode"]
            .apply(lambda x: x.tolist())
            .to_dict()
        )
        df_marker_top5 = df_marker.groupby("cluster").head(5)
        df_marker_top5["celltype"] = None

        target_genes = set(df_marker_top5.gene)
        cell_type = "T_cells" if self.seqtype == "TCR" else "B_cells"
        for gene in target_genes:
            if gene in marker_dict[self.species][cell_type]:
                df_marker_top5.loc[df_marker_top5["gene"] == gene, "celltype"] = (
                    cell_type
                )

        target_clusters = df_marker_top5[
            df_marker_top5["celltype"] == cell_type
        ].cluster.tolist()
        target_clusters = [int(x.split(" ")[-1]) for x in target_clusters]
        target_clusters = set(
            [x for x in target_clusters if target_clusters.count(x) > 1]
        )

        if not target_clusters:
            return None, self.args.expected_target_cell_num

        target_barcodes = []
        for cluster in target_clusters:
            target_barcodes += bc_cluster_dict[cluster]

        return target_barcodes, len(target_barcodes)

    @utils.add_log
    def parse_contig_file(self):
        """
        Convert to all_contig_annotation file in 10X format.
        Generate all_contig_fasta file.
        """
        self.convert_barcode_report(
            self.barcode_report, outdir=f"{self.outdir}/{self.sample}"
        )

        if self.seqtype == "BCR":
            df = pd.read_csv(f"{self.outdir}/{self.sample}_b.csv")
        else:
            df = pd.read_csv(f"{self.outdir}/{self.sample}_t.csv")

        df["productive"] = df["full_length"]
        contig_set = set(df.contig_id)

        # generate all contig fasta file
        # add length of each contig.
        len_dict = dict()
        all_fa = open(f"{self.outdir}/{self.sample}_all_contig.fasta", "w")
        with pysam.FastxFile(self.annot) as fa:
            for read in fa:
                len_dict[read.name] = read.comment.split(" ")[0]
                if read.name in contig_set:
                    sequence = read.sequence
                    all_fa.write(">" + read.name + "\n" + sequence + "\n")
        all_fa.close()
        df["length"] = df["contig_id"].apply(len_dict.get)

        return df

    @utils.add_log
    def cell_calling(self, df):
        """
        Common filtering based on CDR3:
        Filter nonfunctional CDR3(shown 'out_of_frame' in cdr3 report), or CDR3 sequences containing "N" in the nucleotide sequence.
        Keep CDR3aa start with C.
        Keep CDR3aa length >= 5.
        Keep no stop codon in CDR3aa.
        Filter low abundance contigs based on a umi cut-off.

        DiffuseFrac filtering(option --diffuseFrac needed):
        If cell A's two chains CDR3s are identical to another cell B,
        and A's chain abundance is significantly lower than B's (--diffuseFrac), filter A.

        Target cell barcodes filtering(option, --target_cell_barcode needed):
        Filter low abundance contigs based on a umi cut-off.
        The umi counts of all contigs originated from B cells are multiplied by a weight to
        better distinguish signal from background noise.
        """
        df.sort_values(by="umis", ascending=False, inplace=True)
        if self.seqtype == "BCR":
            df_chain_heavy = df[df["chain"] == "IGH"]
            df_chain_light = df[(df["chain"] == "IGK") | (df["chain"] == "IGL")]
        else:
            df_chain_heavy = df[df["chain"] == "TRA"]
            df_chain_light = df[df["chain"] == "TRB"]
        df_chain_heavy = df_chain_heavy.drop_duplicates(["barcode"])
        df_chain_light = df_chain_light.drop_duplicates(["barcode"])
        df_for_clono = pd.concat([df_chain_heavy, df_chain_light], ignore_index=True)

        self.add_cell_num_metric(df_for_clono, "Total Assembled Cell Number")

        # Common filtering
        trust_report = pd.read_csv(self.trust_report, sep="\t")
        correct_cdr3 = set(df_for_clono.cdr3).intersection(set(trust_report.CDR3aa))
        correct_cdr3 = [i for i in correct_cdr3 if i.startswith("C")]
        correct_cdr3 = [i for i in correct_cdr3 if len(i) >= 5]
        correct_cdr3 = [i for i in correct_cdr3 if "UAG" or "UAA" or "UGA" not in i]
        df_for_clono = df_for_clono[df_for_clono["cdr3"].isin(correct_cdr3)]

        self.add_cell_num_metric(df_for_clono, "Cell Number after CDR3 filtering")

        # Filter low abundance contigs based on a umi cut-off
        if self.seqtype == "BCR":
            df_chain_heavy = df_for_clono[df_for_clono["chain"] == "IGH"]
            df_chain_light = df_for_clono[
                (df_for_clono["chain"] == "IGK") | (df_for_clono["chain"] == "IGL")
            ]
        else:
            df_chain_heavy = df_for_clono[df_for_clono["chain"] == "TRA"]
            df_chain_light = df_for_clono[df_for_clono["chain"] == "TRB"]

        filtered_congtigs_id = set()
        for _df in [df_chain_heavy, df_chain_light]:
            target_contigs = target_cell_calling(
                _df,
                expected_target_cell_num=self.expected_target_cell_num,
                target_barcodes=self.target_barcodes,
                weight=self.target_weight,
                coef=self.coef,
            )
            filtered_congtigs_id = filtered_congtigs_id | target_contigs

        df_for_clono = df_for_clono[df_for_clono.contig_id.isin(filtered_congtigs_id)]
        self.add_cell_num_metric(df_for_clono, "Cell Number after UMI filtering")

        df_for_clono_pro = df_for_clono[df_for_clono["productive"]]
        cell_barcodes, filtered_contig = (
            set(df_for_clono_pro["barcode"]),
            set(df_for_clono_pro["contig_id"]),
        )

        return df_for_clono, cell_barcodes, filtered_contig

    @utils.add_log
    def filter_fasta(self, cell_barcodes):
        """Filter all contig fasta file by barcodes which are identified to be cell.

        :param cell_barcodes: all barcodes identified to be cell.
        """
        all_contig_fasta = f"{self.outdir}/{self.sample}_all_contig.fasta"
        filter_contig_fasta = f"{self.outdir}/{self.sample}_filtered_contig.fasta"

        filter_contig_fasta = open(filter_contig_fasta, "w")
        with pysam.FastxFile(all_contig_fasta) as fa:
            for read in fa:
                name = read.name
                cb = "_".join(
                    name.split("_")[:-1]
                )  # remove the suffix num in "bc1_bc2_bc3_num"
                sequence = read.sequence
                if cb in cell_barcodes:
                    filter_contig_fasta.write(">" + name + "\n" + sequence + "\n")

        filter_contig_fasta.close()

    @utils.add_log
    def parse_clonotypes(self, df, df_for_clono, cell_barcodes, filtered_contig):
        """Parse clonotypes from CDR3 and manually add clonotype id for each contig.

        :param df: original contig file.
        :param df_for_clono: contig info after filter.
        :param cell_barcodes: all barcodes identified to be cell.
        :return df_filter_contig: filtered contigs by cell barcodes.
        """
        df_for_clono_pro = df_for_clono[df_for_clono["productive"]].copy()
        df_for_clono_pro["chain_cdr3aa"] = df_for_clono_pro.loc[
            :, ["chain", "cdr3"]
        ].apply(":".join, axis=1)
        df_for_clono_pro["chain_cdr3nt"] = df_for_clono_pro.loc[
            :, ["chain", "cdr3_nt"]
        ].apply(":".join, axis=1)

        cbs = set(df_for_clono_pro["barcode"])
        clonotypes = open(f"{self.outdir}/clonotypes.csv", "w")
        clonotypes.write("barcode\tcdr3s_aa\tcdr3s_nt\n")
        for cb in cbs:
            temp = df_for_clono_pro[df_for_clono_pro["barcode"] == cb]
            temp = temp.sort_values(by="chain", ascending=True)
            aa_chain = ";".join(list(temp["chain_cdr3aa"]))
            nt_chain = ";".join(list(temp["chain_cdr3nt"]))
            clonotypes.write(f"{cb}\t{aa_chain}\t{nt_chain}\n")
        clonotypes.close()

        df_clonotypes = pd.read_csv(
            f"{self.outdir}/clonotypes.csv", sep="\t", index_col=None
        )
        contig_with_clonotype = copy.deepcopy(df_clonotypes)
        df_dict = (
            df_clonotypes[["cdr3s_nt", "cdr3s_aa"]]
            .set_index("cdr3s_nt")
            .to_dict(orient="dict")["cdr3s_aa"]
        )
        df_clonotypes = df_clonotypes.groupby("cdr3s_nt", as_index=False).agg(
            {"barcode": "count"}
        )
        df_clonotypes.rename(columns={"barcode": "frequency"}, inplace=True)
        sum_f = df_clonotypes["frequency"].sum()

        df_clonotypes["proportion"] = df_clonotypes["frequency"].apply(
            lambda x: x / sum_f
        )
        df_clonotypes.sort_values(by="frequency", ascending=False, inplace=True)
        df_clonotypes["clonotype_id"] = [
            f"clonotype{i}" for i in range(1, df_clonotypes.shape[0] + 1)
        ]
        df_clonotypes["cdr3s_aa"] = df_clonotypes["cdr3s_nt"].apply(
            lambda x: df_dict[x]
        )
        df_clonotypes = df_clonotypes.reindex(
            columns=["clonotype_id", "frequency", "proportion", "cdr3s_aa", "cdr3s_nt"]
        )
        df_clonotypes.to_csv(f"{self.outdir}/clonotypes.csv", sep=",", index=False)
        used_for_merge = df_clonotypes[["cdr3s_nt", "clonotype_id"]]

        df_merge = pd.merge(
            used_for_merge, contig_with_clonotype, on="cdr3s_nt", how="outer"
        )
        df_merge = df_merge[["barcode", "clonotype_id"]]
        df_all_contig = pd.merge(df_merge, df, on="barcode", how="outer")
        df_all_contig.fillna("", inplace=True)
        df_all_contig = df_all_contig[
            [
                "barcode",
                "is_cell",
                "contig_id",
                "high_confidence",
                "length",
                "chain",
                "v_gene",
                "d_gene",
                "j_gene",
                "c_gene",
                "full_length",
                "productive",
                "cdr3",
                "cdr3_nt",
                "reads",
                "umis",
                "clonotype_id",
            ]
        ]
        df_filter_contig = df_all_contig[df_all_contig["barcode"].isin(cell_barcodes)]
        for _df in [df_all_contig, df_filter_contig]:
            _df.loc[~_df.contig_id.isin(filtered_contig), "clonotype_id"] = ""

        df_all_contig.to_csv(
            f"{self.outdir}/{self.sample}_all_contig.csv", sep=",", index=False
        )
        df_filter_contig.to_csv(
            f"{self.outdir}/{self.sample}_filtered_contig.csv", sep=",", index=False
        )

    @utils.add_log
    def gen_summary(self, df_for_clono):
        """Generate metrics in html"""
        df_for_clono_pro = df_for_clono[df_for_clono["productive"]]
        cell_barcodes = set(df_for_clono_pro["barcode"])
        total_cells = len(cell_barcodes)

        read_count = 0
        read_count_all = 0
        umi_dict = defaultdict(set)
        umi_count = defaultdict()
        with pysam.FastxFile(self.fq2) as fq:
            for read in fq:
                read_count_all += 1
                cb = read.name.split(":")[0]
                umi = read.name.split(":")[1]
                umi_dict[cb].add(umi)
                if cb in cell_barcodes:
                    read_count += 1
        for cb in umi_dict:
            umi_count[cb] = len(umi_dict[cb])
        df_umi = pd.DataFrame.from_dict(umi_count, orient="index", columns=["UMI"])
        df_umi["barcode"] = df_umi.index
        df_umi = df_umi.reset_index(drop=True)
        df_umi = df_umi.reindex(columns=["barcode", "UMI"])
        df_umi = df_umi.sort_values(by="UMI", ascending=False)
        df_umi["mark"] = df_umi["barcode"].apply(
            lambda x: "CB" if x in cell_barcodes else "UB"
        )
        df_umi.to_csv(f"{self.outdir}/count.txt", sep="\t", index=False)
        self.add_data(
            chart=get_plot_elements.plot_barcode_rank(f"{self.outdir}/count.txt")
        )

        self.add_metric(
            name="Estimated Number of Cells",
            value=total_cells,
            help_info="Number of cells which contain at least one chain (for TCR: TRA or TRB, for BCR: IGH, IGL or IGK)",
        )

        self.add_metric(
            name="Mean Read Pairs per Cell",
            value=int(read_count / total_cells),
            help_info="Number of input reads divided by the estimated number of cells",
        )

        used_read = 0
        with pysam.FastxFile(self.assembled_fa) as fa:
            for read in fa:
                bc = read.name.split(":")[0]
                if bc in cell_barcodes:
                    used_read += 1
        self.add_metric(
            name="Mean Used Read Pairs per Cell",
            value=int(used_read / total_cells),
            help_info="Mean number of reads used in assembly per cell-associated barcode",
        )
        self.add_metric(
            name="Fraction of Reads in Cells",
            value=used_read,
            total=read_count_all,
            help_info="Number of reads with cell-associated barcodes divided by the number of reads with valid barcodes",
        )

        for c in self.chains:
            temp_df = df_for_clono_pro[df_for_clono_pro["chain"] == c]

            try:
                median_umi = int(temp_df["umis"].median())
            except ValueError:
                # ValueError: cannot convert float NaN to integer
                median_umi = 0

            self.add_metric(name=f"Median {c} UMIs per Cell", value=median_umi)

    def run(self):
        original_df = self.parse_contig_file()
        df_for_clono, cell_barcodes, filtered_contig = self.cell_calling(original_df)
        self.filter_fasta(cell_barcodes)
        self.parse_clonotypes(original_df, df_for_clono, cell_barcodes, filtered_contig)
        self.gen_summary(df_for_clono)


@utils.add_log
def summarize(args):
    with Summarize(args, display_title="Cells") as runner:
        runner.run()


def get_opts_summarize(parser, sub_program):
    parser.add_argument(
        "--seqtype", help="TCR or BCR", choices=["TCR", "BCR"], required=True
    )
    parser.add_argument(
        "--species",
        help="human or mouse. If not provided, must specify --ref.",
        choices=["human", "mouse"],
    )
    parser.add_argument("--coef", help="coef for auto filter", default=5)
    parser.add_argument(
        "--diffuseFrac",
        help="If cell A's two chains CDR3s are identical to another cell B, and A's chain abundance is significantly lower than B's, filter A.",
        action="store_true",
    )
    parser.add_argument(
        "--expected_target_cell_num",
        help="Expected T or B cell number. If `--target_cell_barcode` is provided, this argument is ignored.",
        type=int,
        default=3000,
    )
    parser.add_argument(
        "--target_cell_barcode",
        help="Barcode of target cells. Auto or path of plain text file with one barcode per line",
        default=None,
    )
    parser.add_argument(
        "--target_weight",
        help="UMIs of the target cells are multiplied by this factor. Only used when `--target_cell_barcode` is provided.",
        type=float,
        default=6.0,
    )
    if sub_program:
        parser = s_common(parser)
        parser.add_argument("--fq2", help="Barcode R2 reads.", required=True)
        parser.add_argument(
            "--assemble_out", help="Result of  assemble dirctory.", required=True
        )
        parser.add_argument(
            "--match_dir", help="Match scRNA-seq directory.", required=True
        )



import os
import sys
import subprocess
from collections import Counter

import pandas as pd
import pysam
import celescope.tools.parse_chemistry as parse_chemistry
from celescope.chemistry_dict import chemistry_dict
from typing import Union
from celescope.tools.__init__ import (
    FILTERED_MATRIX_DIR_SUFFIX,
    COUNTS_FILE_NAME,
)
from celescope.__init__ import HELP_DICT
from celescope.tools.step import Step, s_common
from celescope.tools.barcode import Barcode
from celescope.tools import utils
from celescope.tools.make_ref import MakeRef
from celescope.tools.matrix import CountMatrix
from celescope.tools.emptydrop_cr import get_plot_elements
from celescope.tools.cells import Cells_metrics

SAM_attributes = "NH HI nM AS CR UR CB UB GX GN "
MIN_CELL = 500
MAX_CELL = 60000
MIN_BEAD = 100000
MAX_BEAD = 300000


def create_pattern_args(pattern_dict: dict) -> str:
    """Create starsolo args relate to pattern"""
    if len(pattern_dict["U"]) != 1:
        raise ValueError(
            f"Error: Wrong pattern:{pattern_dict}. \n Solution: fix pattern so that UMI only have 1 position.\n"
        )
    ul = pattern_dict["U"][0].start
    ur = pattern_dict["U"][0].stop
    umi_len = ur - ul

    if len(pattern_dict["C"]) == 1:
        solo_type = "CB_UMI_Simple"
        start, stop = pattern_dict["C"][0].start, pattern_dict["C"][0].stop
        cb_start = start + 1
        cb_len = stop - start
        umi_start = ul + 1
        cb_str = f"--soloCBstart {cb_start} --soloCBlen {cb_len} "
        umi_str = f"--soloUMIstart {umi_start} --soloUMIlen {umi_len} "
    else:
        solo_type = "CB_UMI_Complex"
        cb_pos = " ".join([f"0_{x.start}_0_{x.stop-1}" for x in pattern_dict["C"]])
        umi_pos = f"0_{ul}_0_{ur-1}"
        cb_str = f"--soloCBposition {cb_pos} "
        umi_str = f"--soloUMIposition {umi_pos} --soloUMIlen {umi_len} "

    return " ".join([f"--soloType {solo_type} ", cb_str, umi_str])


def create_v3_pattern_args() -> str:
    """Can also be used to create flv_rna-V2
    https://github.com/alexdobin/STAR/issues/1607#issuecomment-1210096785
    2: adapter start
    3: adapter end
    """
    linker1 = "ACGATG"
    linker2 = "CATAGT"
    bc = "N" * 9
    linker_len = 6
    bc2_start = 9 + linker_len
    pattern_args = (
        "--soloType CB_UMI_Complex "
        f"--soloCBposition 2_0_2_8 2_{bc2_start}_2_{bc2_start+8} 3_1_3_9 "
        "--soloUMIposition 3_10_3_21 "
        f"--soloAdapterSequence {bc}{linker1}{bc}{linker2} "
        "--soloAdapterMismatchesNmax 1 "
    )
    return pattern_args


def create_whitelist_args(whitelist_str) -> str:
    if not whitelist_str:
        res = "None"
    else:
        res = whitelist_str.strip()
        # nextflow copy remote file to current folder, so only keep file name.
        if res.startswith("http"):
            res = res.split("/")[-1]
        if res.endswith(".gz"):
            res = f"<(gzip -cdf {res})"
    res = f" --soloCBwhitelist {res} "
    return res


def create_solo_args(
    pattern_args: str,
    whitelist_args: str,
    outFileNamePrefix: str,
    fq1: str,
    fq2: str,
    genomeDir: str,
    soloCellFilter: str,
    runThreadN: Union[str, int],
    clip3pAdapterSeq: str,
    outFilterMatchNmin: Union[str, int],
    soloFeatures: str,
    outSAMtype: str,
    outSAMattributes: str,
    soloCBmatchWLtype: str,
    limitBAMsortRAM: Union[str, int],
    extra_starsolo_args: str,
) -> str:
    """Create all starsolo args"""
    read_command = "zcat" if fq1.strip().endswith(".gz") else "cat"
    cmd = (
        "STAR \\\n"
        f"{pattern_args} \\\n"
        f"{whitelist_args} \\\n"
        f"--outFileNamePrefix {outFileNamePrefix} \\\n"
        f"--readFilesIn {fq2} {fq1} \\\n"
        f"--readFilesCommand {read_command} \\\n"
        f"--genomeDir {genomeDir} \\\n"
        f"--soloCellFilter {soloCellFilter} \\\n"
        f"--runThreadN {runThreadN} \\\n"
        f"--clip3pAdapterSeq {clip3pAdapterSeq} \\\n"
        f"--outFilterMatchNmin {outFilterMatchNmin} \\\n"
        f"--soloFeatures {soloFeatures} \\\n"
        f"--soloCBmatchWLtype {soloCBmatchWLtype} \\\n"
        f"--limitBAMsortRAM {limitBAMsortRAM} \\\n"
        f"--outSAMtype {outSAMtype} \\\n"
        "--soloCellReadStats Standard \\\n"
        "--soloBarcodeReadLength 0 \\\n"
        f"{extra_starsolo_args} \\\n"
    )
    if outSAMtype != "None":
        cmd += f"--outSAMattributes {outSAMattributes} \\\n"
    return cmd


def create_soloFeatures(args_soloFeatures, report_soloFeature):
    """
    BAM uses first feature in soloFeatures, so move report_soloFeature to the first position.
    GeneFull_Ex50pAS is also need for intron region metrics.
    """
    features = args_soloFeatures.split(" ")
    if report_soloFeature in features:
        features.remove(report_soloFeature)
    features = [report_soloFeature] + features
    if "GeneFull_Ex50pAS" not in features:
        features.append("GeneFull_Ex50pAS")
    return " ".join(features)


class Starsolo(Step):
    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)
        self.fq1_list = args.fq1.split(",")
        self.fq2_list = args.fq2.split(",")
        if len(self.fq1_list) != len(self.fq2_list):
            sys.exit("fq1 and fq2 must have same number of files")

        self.chemistry = parse_chemistry.get_chemistry(
            self.assay, args.chemistry, self.fq1_list
        )
        self.pattern_dict, self.bc = parse_chemistry.get_pattern_dict_and_bc(
            self.chemistry, args.pattern, args.whitelist
        )

        self.pattern_args = (
            create_pattern_args(self.pattern_dict)
            if self.chemistry not in ("GEXSCOPE-V3", "flv_rna-V2")
            else create_v3_pattern_args()
        )
        whitelist_str = " ".join(self.bc)
        self.whitelist_args = create_whitelist_args(whitelist_str)
        self.outSAMattributes = SAM_attributes + self.args.SAM_attributes
        self.extra_starsolo_args = args.STAR_param

        # output files
        self.solo_out_dir = f"{self.outdir}/{self.sample}_Solo.out/"
        solo_dir = f"{self.outdir}/{self.sample}_Solo.out/{args.report_soloFeature}"
        self.raw_matrix = f"{solo_dir}/raw"
        self.filtered_matrix = f"{solo_dir}/filtered"
        self.summary_file = f"{solo_dir}/Summary.csv"
        bam = f"{self.outdir}/{self.sample}_Aligned.sortedByCoord.out.bam"

        # outs
        self.outs = [self.raw_matrix, self.filtered_matrix, bam]

    def run_starsolo(self):
        soloFeatures = create_soloFeatures(
            self.args.soloFeatures, self.args.report_soloFeature
        )
        cmd = create_solo_args(
            pattern_args=self.pattern_args,
            whitelist_args=self.whitelist_args,
            outFileNamePrefix=self.out_prefix + "_",
            fq1=self.args.fq1,
            fq2=self.args.fq2,
            genomeDir=self.args.genomeDir,
            soloCellFilter=self.args.soloCellFilter,
            runThreadN=self.args.thread,
            clip3pAdapterSeq=self.args.adapter_3p,
            outFilterMatchNmin=self.args.outFilterMatchNmin,
            soloFeatures=soloFeatures,
            outSAMtype=self.args.outSAMtype,
            outSAMattributes=self.outSAMattributes,
            soloCBmatchWLtype=self.args.soloCBmatchWLtype,
            limitBAMsortRAM=self.args.limitBAMsortRAM,
            extra_starsolo_args=self.extra_starsolo_args,
        )
        sys.stderr.write(cmd)
        subprocess.check_call(cmd, shell=True)
        cmd = f"chmod -R 755 {self.solo_out_dir}"
        sys.stderr.write(cmd)
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def gzip_matrix(self):
        if not os.path.exists(self.filtered_matrix):
            sys.exit(
                f"{self.filtered_matrix} not found. This error indicates that StarSolo did not generate the gene expression matrix correctly. The specific cause of this error can usually be found in the log above. A common cause of the error is that the fastq file was truncated due to incomplete download."
            )
        cmd = f"gzip {self.raw_matrix}/*; gzip {self.filtered_matrix}/*"
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def get_Q30_cb_UMI(self):
        fq1_list = self.args.fq1.split(",")
        pattern_dict = self.pattern_dict
        cb_10k, umi_10k, cb_10k_100k, umi_10k_100k = (
            Counter(),
            Counter(),
            Counter(),
            Counter(),
        )
        n = 0
        with pysam.FastxFile(fq1_list[0], persist=False) as fq1:
            for entry in fq1:
                n += 1
                if n > 10**5:
                    break
                qual: str = entry.quality
                cb_qual = "".join([qual[slice] for slice in pattern_dict["C"]])
                umi_qual = "".join([qual[slice] for slice in pattern_dict["U"]])
                if n <= 10**4:
                    cb_10k.update(cb_qual)
                    umi_10k.update(umi_qual)
                else:
                    cb_10k_100k.update(cb_qual)
                    umi_10k_100k.update(umi_qual)

        cb_qual_counter = cb_10k
        umi_qual_counter = umi_10k
        if cb_10k_100k:
            cb_qual_counter = cb_10k_100k
            umi_qual_counter = umi_10k_100k
        q30_cb = sum(
            [cb_qual_counter[k] for k in cb_qual_counter if Barcode.chr_to_int(k) >= 30]
        ) / float(sum(cb_qual_counter.values()))
        q30_umi = sum(
            [
                umi_qual_counter[k]
                for k in umi_qual_counter
                if Barcode.chr_to_int(k) >= 30
            ]
        ) / float(sum(umi_qual_counter.values()))
        return q30_cb, q30_umi

    def run(self):
        self.run_starsolo()
        self.gzip_matrix()
        q30_cb, q30_umi = self.get_Q30_cb_UMI()
        return q30_cb, q30_umi, self.chemistry


def starsolo(args):
    with Starsolo(args) as runner:
        q30_cb, q30_umi, chemistry = runner.run()

    with Mapping(args) as runner:
        valid_reads, corrected = runner.run()

    with Cells(args) as runner:
        n_reads, q30_RNA = runner.run(chemistry, valid_reads)

    with Demultiplexing(args) as runner:
        runner.run(valid_reads, n_reads, corrected, q30_cb, q30_umi, q30_RNA)


class Mapping(Step):
    # only add metrics
    def __init__(self, args, display_title=None):
        super().__init__(args, display_title=display_title)
        solo_dir = f"{self.outdir}/{self.sample}_Solo.out/GeneFull_Ex50pAS"
        self.cellReadsStats = f"{solo_dir}/CellReads.stats"
        self.filtered_matrix = f"{self.outs_dir}/{FILTERED_MATRIX_DIR_SUFFIX}"
        self.counts_file = f"{solo_dir}/{COUNTS_FILE_NAME }"
        self.genome = MakeRef.get_config(args.genomeDir)["meta"]["genome_name"]

        self.outs = [self.counts_file]

    @utils.add_log
    def parse_cellReadsStats(self) -> dict[str, int]:
        df = pd.read_csv(self.cellReadsStats, sep="\t", header=0, index_col=0)
        df = df.iloc[1:,]  # skip first line cb not pass whitelist
        df_count = df.loc[:, ["nUMIunique", "countedU"]]  # keep dataframe format
        df_count.rename(columns={"nUMIunique": "UMI"}, inplace=True)
        df_count.sort_values(by="UMI", ascending=False, inplace=True)
        cbs = CountMatrix.read_barcodes(self.filtered_matrix)
        df_count["mark"] = "UB"
        for cb in cbs:
            df_count.loc[cb, "mark"] = "CB"
        df_count.to_csv(self.counts_file, sep="\t", index=True)

        df = df.loc[
            :,
            [
                "cbMatch",
                "cbPerfect",
                "genomeU",
                "genomeM",
                "exonic",
                "intronic",
                "exonicAS",
                "intronicAS",
                "countedU",
            ],
        ]
        metrics = df.sum().to_dict()
        return metrics

    @utils.add_log
    def add_metrics_to_report(self, metrics: dict[str, int]):
        valid = metrics["cbMatch"]
        perfect = metrics["cbPerfect"]
        corrected = valid - perfect
        genomeU = metrics["genomeU"]
        genomeM = metrics["genomeM"]
        mapped = genomeU + genomeM
        exonic = metrics["exonic"]
        intronic = metrics["intronic"]
        antisense = metrics["exonicAS"] + metrics["intronicAS"]
        intergenic = mapped - exonic - intronic - antisense
        countedU = metrics["countedU"]

        self.add_metric(
            name="genome",
            value=self.genome,
        )

        self.add_metric(
            name="Reads mapped to unique loci",
            value=genomeU / valid,
            value_type="fraction",
            help_info="Reads that mapped uniquely to the genome.",
        )

        self.add_metric(
            name="Reads mapped to multiple loci",
            value=genomeM / valid,
            value_type="fraction",
            help_info="Reads that mapped to multiple loci in the genome",
        )
        unique_transcriptome = countedU / valid
        self.add_metric(
            name="Reads mapped uniquely to Transcriptome",
            value=unique_transcriptome,
            value_type="fraction",
            help_info="Reads that mapped to a unique gene in the transcriptome. These reads are used for UMI counting.",
        )
        self.add_metric(
            name="Mapped Reads assigned to exonic regions",
            value=exonic / mapped,
            value_type="fraction",
            help_info="Reads that assigned to exonic regions of genes",
        )
        self.add_metric(
            name="Mapped Reads assigned to intronic regions",
            value=intronic / mapped,
            value_type="fraction",
            help_info="Reads that assigned to intronic regions of genes",
        )
        self.add_metric(
            name="Mapped Reads assigned to intergenic regions",
            value=intergenic / mapped,
            value_type="fraction",
            help_info="Reads that can not be assigned to a gene will be considered as intergenic reads.",
        )
        self.add_metric(
            name="Mapped Reads assigned Antisense to gene",
            value=antisense / mapped,
            value_type="fraction",
            help_info="Reads that assigned to the opposite strand of genes",
        )
        return valid, corrected

    @utils.add_log
    def run(self):
        metrics = self.parse_cellReadsStats()
        return self.add_metrics_to_report(metrics)


class Cells(Cells_metrics):
    def __init__(self, args, display_title=None):
        super().__init__(args, display_title=display_title)
        solo_dir = f"{self.outdir}/{self.sample}_Solo.out/{args.report_soloFeature}"
        self.summary_file = f"{solo_dir}/Summary.csv"
        self.counts_file = f"{self.outs_dir}/{COUNTS_FILE_NAME}"

    @utils.add_log
    def parse_summary_add_metrics(self, valid_reads):
        df = pd.read_csv(self.summary_file, index_col=0, header=None)
        s = df.iloc[:, 0]
        n_cells = int(s["Estimated Number of Cells"])
        fraction_reads_in_cells = float(s["Fraction of Unique Reads in Cells"])
        mean_used_reads_per_cell = int(s["Mean Reads per Cell"])
        median_umi_per_cell = int(s["Median UMI per Cell"])
        median_genes_per_cell = int(
            s[f"Median {self.args.report_soloFeature} per Cell"]
        )
        total_genes = int(s[f"Total {self.args.report_soloFeature} Detected"])
        saturation = float(s["Sequencing Saturation"])
        n_reads = int(s["Number of Reads"])
        q30_RNA = float(s["Q30 Bases in RNA read"])

        self.add_cells_metrics(
            n_cells,
            fraction_reads_in_cells,
            mean_used_reads_per_cell,
            median_umi_per_cell,
            total_genes,
            median_genes_per_cell,
            saturation,
            valid_reads,
        )
        return n_reads, q30_RNA, fraction_reads_in_cells

    def run(self, chemistry, valid_reads):
        n_reads, q30_RNA, fraction_reads_in_cells = self.parse_summary_add_metrics(
            valid_reads
        )
        self.add_data(chart=get_plot_elements.plot_barcode_rank(self.counts_file))
        return n_reads, q30_RNA


class Demultiplexing(Step):
    def __init__(self, args, display_title=None):
        super().__init__(args, display_title=display_title)

    def run(self, valid_reads, n_reads, corrected, q30_cb, q30_umi, q30_RNA):
        self.add_metric(
            name="Raw Reads", value=n_reads, help_info="total reads from FASTQ files"
        )
        self.add_metric(
            name="Valid Reads",
            value=valid_reads / n_reads,
            value_type="fraction",
            help_info="fraction of reads with valid barcode and UMI",
        )

        self.add_metric(
            name="Corrected Barcodes",
            value=corrected / valid_reads,
            value_type="fraction",
            help_info="fraction of valid reads with corrected barcodes. Barcodes are corrected to the whitelist sequence that is 1 Hamming-distance away.",
        )

        self.add_metric(
            name="Q30 of Barcodes",
            value=q30_cb,
            value_type="fraction",
            help_info="Fraction of barcode bases with quality score >= 30",
        )

        self.add_metric(
            name="Q30 of UMI",
            value=q30_umi,
            value_type="fraction",
            help_info="Fraction of UMI bases with quality score >= 30",
        )

        self.add_metric(
            name="Q30 of RNA Reads",
            value=q30_RNA,
            value_type="fraction",
            help_info="Fraction of RNA read bases with quality score >= 30",
        )


def get_opts_starsolo(parser, sub_program=True):
    parser.add_argument(
        "--chemistry",
        help=HELP_DICT["chemistry"],
        choices=list(chemistry_dict.keys()),
        default="auto",
    )
    parser.add_argument(
        "--pattern",
        help="""The pattern of R1 reads, e.g. `C8L16C8L16C8L1U12T18`. The number after the letter represents the number 
        of bases.  
        - `C`: cell barcode  
        - `L`: linker(common sequences)  
        - `U`: UMI    
        - `T`: poly T""",
    )
    parser.add_argument(
        "--whitelist",
        help="Cell barcode whitelist file path, one cell barcode per line.",
    )
    parser.add_argument(
        "--adapter_3p",
        help="Adapter sequence to clip from 3 prime. Multiple sequences are seperated by space",
        default="AAAAAAAAAAAA",
    )
    parser.add_argument(
        "--genomeDir",
        help=HELP_DICT["genomeDir"],
    )
    parser.add_argument(
        "--outFilterMatchNmin",
        help="""Alignment will be output only if the number of matched bases 
is higher than or equal to this value.""",
        default=50,
    )
    parser.add_argument(
        "--soloCellFilter",
        help="Same as the argument in STARsolo",
        default="EmptyDrops_CR 3000 0.99 10 45000 90000 500 0.01 20000 0.001 10000",
    )
    parser.add_argument(
        "--limitBAMsortRAM",
        help="Same as the argument in STARsolo",
        default=32000000000,
        type=int,
    )
    parser.add_argument("--STAR_param", help=HELP_DICT["additional_param"], default="")
    parser.add_argument(
        "--outSAMtype",
        help="Same as the argument in STARsolo. Set to 'None' to skip BAM file output.",
        default="BAM SortedByCoordinate",
        type=str,
    )
    parser.add_argument(
        "--SAM_attributes",
        help=f"Additional attributes(other than {SAM_attributes}) to be added to SAM file",
        default="",
    )
    parser.add_argument(
        "--soloFeatures",
        help="Same as the argument in STARsolo",
        default="GeneFull_Ex50pAS Gene",
    )
    parser.add_argument(
        "--soloCBmatchWLtype",
        help="Same as the argument in STARsolo. Please note `EditDist 2` only works with `--soloType CB UMI Complex`. ",
        default="1MM",
    )
    parser.add_argument(
        "--report_soloFeature",
        help="Specify which soloFeatures to use in the HTML report and the outs directory.",
        default="GeneFull_Ex50pAS",
    )

    if sub_program:
        parser.add_argument(
            "--fq1",
            help="R1 fastq file. Multiple files are separated by comma.",
            required=True,
        )
        parser.add_argument(
            "--fq2",
            help="R2 fastq file. Multiple files are separated by comma.",
            required=True,
        )
        parser = s_common(parser)

    return parser



import os
import unittest
from collections import namedtuple

from celescope.tools.step import Step


class Tests(unittest.TestCase):
    def setUp(self):
        pass

    @unittest.skip("tested")
    def test_stat_to_metric(self):
        os.chdir("/SGRNJ01/RD_dir/pipeline_test/zhouyiqi/multi_tests/rna")
        args_dict = {
            "sample": "test1",
            "assay": "rna",
            "thread": 1,
            "outdir": "test1/06.analysis",
            "debug": True,
        }
        Args = namedtuple("Args", list(args_dict.keys()))
        args = Args(**args_dict)

        obj = Step(args, "analysis")
        obj.stat_to_metric()
        print(obj.__content_dict["metric"])

    def test_test(self):
        assert 0 == 0



from celescope.tools.starsolo import (
    Starsolo,
    Mapping,
    Cells,
    Demultiplexing,
    get_opts_starsolo as opts_super,
)


def starsolo(args):
    args.STAR_param += " --outSAMunmapped Within "
    with Starsolo(args) as runner:
        q30_cb, q30_umi, chemistry = runner.run()
    with Mapping(args, display_title="Mapping to Host Genome") as runner:
        valid_reads, corrected = runner.run()
    with Cells(args) as runner:
        n_reads, q30_RNA = runner.run(chemistry, valid_reads)
    with Demultiplexing(args) as runner:
        runner.run(valid_reads, n_reads, corrected, q30_cb, q30_umi, q30_RNA)


def get_opts_starsolo(parser, sub_program):
    opts_super(parser, sub_program)



# Li Song: the program to re-cluster the CDR3 regions

#!/usr/bin/env python3

import sys
import math


def GetChainType(v, j, c):
    s = ""
    if v != "*":
        s = v
    elif c != "*":
        s = c
    elif j != "*":
        s = j
    else:
        return -1

    if s[0:3] == "IGH":
        return 0
    elif s[0:3] == "IGK":
        return 1
    elif s[0:3] == "IGL":
        return 2
    elif s[0:3] == "TRA":
        return 3
    elif s[0:3] == "TRB":
        return 4
    elif s[0:3] == "TRG":
        return 5
    elif s[0:3] == "TRD":
        return 6
    else:
        return -1


def GetMainGeneName(g):
    return g.split("*")[0]


def CompatibleGeneAssignment(a, b):
    if GetChainType(a[0], a[2], a[3]) != GetChainType(b[0], b[2], b[3]):
        return False

    for i in [0, 2]:  # V, J gene assignment
        ga = a[i].split("*")[0]
        gb = b[i].split("*")[0]
        if ga != "" and gb != "" and ga != gb:
            return False
    return True


def GetSimilarity(a, b):
    if len(a) != len(b):
        return 0
    sameCnt = 0
    for i in range(len(a)):
        if a[i] == b[i]:
            sameCnt += 1
    return sameCnt / len(a)


def CompatibleSequence(a, b, similarity):
    if len(a) != len(b):
        return False
    diffCnt = 0
    diffMax = len(a) - int(math.ceil(len(a) * similarity))
    for i in range(len(a)):
        if a[i] != b[i]:
            diffCnt += 1
            if diffCnt > diffMax:
                return False
    return True


def GetFather(tag, father):
    if father[tag] != tag:
        father[tag] = GetFather(father[tag], father)
    return father[tag]


def LargerCluster(rawCdr3List, similarity, prefix, useRepresentative, mode):
    vjCDR3LenList = {}
    clusterNameToId = {}
    clusterIdToName = []
    clusterRepresentativeId = {}
    clusterRepresentativeAbund = {}

    if len(rawCdr3List) == 0:
        return
    cdr3List = sorted(rawCdr3List, key=lambda x: (x[0], x[8]))
    # Select the represenative, allowing same CDR3 shown up in different rows in a cluster to handle scRNA-seq
    abund = cdr3List[0][10]
    for i in range(1, len(cdr3List) + 1):
        prevKey = (cdr3List[i - 1][0], cdr3List[i - 1][8])
        key = "*"
        if i < len(cdr3List):
            key = (cdr3List[i][0], cdr3List[i][8])
        if key == prevKey:
            abund += cdr3List[i][10]
        else:
            cdr3 = cdr3List[i - 1]
            if cdr3[0] not in clusterNameToId:
                clusterNameToId[cdr3[0]] = len(clusterIdToName)
                clusterIdToName.append(cdr3[0])
                clusterRepresentativeId[cdr3[0]] = i - 1
                clusterRepresentativeAbund[cdr3[0]] = abund
            elif abund > clusterRepresentativeAbund[cdr3[0]]:
                clusterRepresentativeId[cdr3[0]] = i - 1
                clusterRepresentativeAbund[cdr3[0]] = abund
            if i < len(cdr3List):
                abund = cdr3List[i][10]
    # print(cdr3List[clusterRepresentativeId["assemble2"]])
    # Reorganize the CDR3 list by their V, J and CDR3 length.
    i = 0
    for cdr3 in cdr3List:
        if useRepresentative and clusterRepresentativeId[cdr3[0]] != i:
            i += 1
            continue
        key = (GetMainGeneName(cdr3[2]), GetMainGeneName(cdr3[4]), len(cdr3[8]))
        if key not in vjCDR3LenList:
            vjCDR3LenList[key] = []
        vjCDR3LenList[key].append(i)
        i += 1
    # Prepare the set-union
    father = []
    for cdr3 in cdr3List:
        father.append(clusterRepresentativeId[cdr3[0]])

    # Build up the set-union relation
    if mode == "aggressive":
        for key in vjCDR3LenList.keys():
            cdr3IdList = vjCDR3LenList[key]
            size = len(cdr3IdList)
            for i in range(size):
                fi = GetFather(cdr3IdList[i], father)
                for j in range(i + 1, size):
                    fj = GetFather(cdr3IdList[j], father)
                    if fi != fj and CompatibleSequence(
                        cdr3List[cdr3IdList[i]][8],
                        cdr3List[cdr3IdList[j]][8],
                        similarity,
                    ):
                        # if (cdr3List[cdr3IdList[j]][8] == "TGCCAACAGTATATTAGTTACTCGTACACTTTT"):
                        # print(i, cdr3List[cdr3IdList[i]])
                        # print(j, cdr3List[cdr3IdList[j]])
                        father[fj] = fi
    elif mode == "center":
        for key in vjCDR3LenList.keys():
            rawCdr3IdList = vjCDR3LenList[key][:]
            cdr3IdList = sorted(
                rawCdr3IdList,
                key=lambda x: (
                    clusterRepresentativeAbund[cdr3List[x][0]],
                    cdr3List[x][10],
                ),
                reverse=True,
            )
            size = len(cdr3IdList)
            for i in range(1, size):
                maxFj = 0
                maxSimilarity = -1
                fi = GetFather(cdr3IdList[i], father)
                for j in range(i):
                    fj = GetFather(cdr3IdList[j], father)
                    if fi == fj:
                        continue
                    s = GetSimilarity(cdr3List[fi], cdr3List[fj])
                    if s > maxSimilarity:
                        maxSimilarity = s
                        maxFj = fj
                if maxSimilarity >= similarity:
                    father[fi] = maxFj

    # Associate original clusters to larger cluster
    largerClusterToId = []
    largerClusterToClusterName = []
    rootToLargerClusterId = {}
    i = 0
    for cdr3 in cdr3List:
        root = GetFather(i, father)
        lcId = 0
        if root not in rootToLargerClusterId:
            rootToLargerClusterId[root] = len(largerClusterToId)
            lcId = len(largerClusterToId)
            largerClusterToId.append([])
            largerClusterToClusterName.append(set({}))
        else:
            lcId = rootToLargerClusterId[root]

        largerClusterToId[lcId].append(i)
        largerClusterToClusterName[lcId].add(cdr3[0])
        i += 1

    # Output the result
    # Output the composition of larger cluster
    # for i in range(len(largerClusterToId)):
    # print("#\tcluster" + str(i), end = "")
    # for name in largerClusterToClusterName[i]:
    # print("\t" + name, end = "")
    # print("\n", end = "")

    # Output the new cdr3 format with new larger cluster Id.
    for i in range(len(largerClusterToId)):
        j = 0
        for cdr3Id in largerClusterToId[i]:
            cdr3List[cdr3Id].append(cdr3List[cdr3Id][0])
            cdr3List[cdr3Id].append(cdr3List[cdr3Id][1])
            cdr3List[cdr3Id][0] = (
                prefix + "_" + str(i)
            )  # + "_" + cdr3List[cdr3Id][0] + "_" + str(cdr3List[cdr3Id][1])
            cdr3List[cdr3Id][1] = j
            print("\t".join(str(x) for x in cdr3List[cdr3Id]))
            j += 1

    return


if __name__ == "__main__":
    if len(sys.argv) <= 1:
        print(
            "usage: a.py trust_cdr3.out [OPTIONS] > output\n"
            + "OPTIONS:\n"
            + "\t-s FLOAT: similarity of two CDR3s (default: 0.8)\n"
            + "\t--prefix STRING: prefix to new cluster name (default: cluster)\n"
            + "\t--center: use the center of the cluster for similarity comparison (default: no)\n"
            + "\t--representative: use representative CDR3 from each contig for cluster (default: no)\n"
            + "\t--format [cdr3, simplerep]: the input format type (default: cdr3)"
        )
        exit(1)

    cdr3List = []
    similarity = 0.8
    prefix = "cluster"
    useRepresentative = False
    mode = "aggressive"
    inputFormat = "cdr3"
    i = 2
    while i < len(sys.argv):
        if sys.argv[i] == "-s":
            similarity = float(sys.argv[i + 1])
            i += 1
        elif sys.argv[i] == "--prefix":
            prefix = sys.argv[i + 1]
            i += 1
        elif sys.argv[i] == "--representative":
            useRepresentative = True
        elif sys.argv[i] == "--center":
            mode = "center"
        elif sys.argv[i] == "--format":
            inputFormat = sys.argv[i + 1]
            if inputFormat not in ["cdr3", "simplerep"]:
                print("Unknown format: ", inputFormat)
                exit(1)
            i += 1
        else:
            print("Unknown option: ", sys.argv[i])
            exit(1)
        i += 1

    fp = open(sys.argv[1])
    lineCnt = 0
    for line in fp:
        line = line.rstrip()
        cols = line.split("\t")
        if inputFormat == "cdr3":
            cols[1] = int(cols[1])
            skip = False
            for g in [2, 4]:  # Must have V, J genes.
                if cols[g] == "*":
                    skip = True
            if float(cols[9]) == 0:
                skip = True
            if skip:
                continue
            for g in [2, 3, 4, 5]:
                cols[g] = cols[g].split(",")[0]
            cols[9] = float(cols[9])
            cols[10] = float(cols[10])
            if cols[9] == 0:
                continue
        elif inputFormat == "simplerep":
            if line[0] == "#":
                continue
            if "_" in cols[3] or "?" in cols[3]:
                continue
            for g in [4, 6]:  # Must have V, J genes.
                if cols[g] == "*":
                    skip = True
            reformat = [0] * 11
            reformat[0] = "line" + str(lineCnt)
            reformat[1] = 0
            for g in [4, 5, 6, 7]:
                reformat[g - 2] = cols[g]
            reformat[6] = reformat[7] = "*"
            reformat[8] = cols[2]
            reformat[9] = 1
            reformat[10] = cols[0]
            cols = reformat[:]
        cdr3List.append(cols)
        lineCnt += 1
    fp.close()
    LargerCluster(cdr3List, similarity, prefix, useRepresentative, mode)



"""
convert 5-prime and 3-prime to the same pattern.
"""

from celescope.tools import utils, parse_chemistry
from celescope.tools.step import Step, s_common
import subprocess
import pysam


class Convert(Step):
    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)

        self.fq1_5p = args.fq1_5p.split(",")
        self.fq1_3p = args.fq1_3p.split(",")
        self.fq2_5p = args.fq2_5p.split(",")
        self.fq2_3p = args.fq2_3p.split(",")

        chemistry_version = args.chemistry.split("-")[1]
        chemistry_5p, chemistry_3p = (
            f"rna_5p-{chemistry_version}",
            f"rna_3p-{chemistry_version}",
        )
        chemistry_dict = parse_chemistry.get_chemistry_dict()
        self.chemistry_dict_5p, self.chemistry_dict_3p = (
            chemistry_dict[chemistry_5p],
            chemistry_dict[chemistry_3p],
        )

    def write_3p(self):
        for i, (fn1, fn2) in enumerate(zip(self.fq1_3p, self.fq2_3p), start=1):
            out_fn1 = f"{self.out_prefix}_3p{i}_R1.fq.gz"
            out_fn2 = f"{self.out_prefix}_3p{i}_R2.fq.gz"
            fh_3p = utils.generic_open(out_fn1, "wt")
            with pysam.FastxFile(fn1, persist=False) as fq:
                for entry1 in fq:
                    name1, seq1, qual1 = entry1.name, entry1.sequence, entry1.quality
                    bc_list, bc_quality_list, umi, umi_quality = (
                        parse_chemistry.get_raw_umi_bc_and_quality(
                            seq1, qual1, self.chemistry_dict_3p["pattern_dict"]
                        )
                    )
                    bc = "".join(bc_list)
                    bc_quality = "".join(bc_quality_list)
                    fh_3p.write(
                        utils.fastq_line(name1, bc + umi, bc_quality + umi_quality)
                    )
            cmd = f"ln -s -f {fn2} {out_fn2}"
            subprocess.check_call(cmd, shell=True)

    def write_5p(self):
        for i, (fn1, fn2) in enumerate(zip(self.fq1_5p, self.fq2_5p), start=1):
            out_fn1 = f"{self.out_prefix}_5p{i}_R1.fq.gz"
            out_fn2 = f"{self.out_prefix}_5p{i}_R2.fq.gz"
            fh_5p_1 = utils.generic_open(out_fn1, "wt")
            fh_5p_2 = utils.generic_open(out_fn2, "wt")
            with pysam.FastxFile(fn1, persist=False) as fq:
                for entry1 in fq:
                    name1, seq1, qual1 = entry1.name, entry1.sequence, entry1.quality
                    bc_list, bc_quality_list, umi, umi_quality = (
                        parse_chemistry.get_raw_umi_bc_and_quality(
                            seq1,
                            qual1,
                            self.chemistry_dict_5p["pattern_dict"],
                            reverse_complement=True,
                        )
                    )
                    bc = "".join(bc_list)
                    bc_quality = "".join(bc_quality_list)
                    fh_5p_1.write(
                        utils.fastq_line(name1, bc + umi, bc_quality + umi_quality)
                    )
            fh_5p_1.close()
            with pysam.FastxFile(fn2, persist=False) as fq:
                for entry2 in fq:
                    name2, seq2, qual2 = entry2.name, entry2.sequence, entry2.quality
                    seq2 = utils.reverse_complement(seq2)
                    qual2 = qual2[::-1]
                    fh_5p_2.write(utils.fastq_line(name2, seq2, qual2))
            fh_5p_2.close()

    def run(self):
        self.write_3p()
        self.write_5p()


@utils.add_log
def convert(args):
    with Convert(args) as runner:
        runner.run()


def get_opts_convert(parser, sub_program=True):
    parser.add_argument(
        "--chemistry",
        required=True,
        choices=["5p3p-1", "5p3p-2", "5p3p-3"],
        help="rna 5p3p chemistry version",
    )
    if sub_program:
        parser.add_argument(
            "--fq1_3p",
            help="3 prime R1 fastq file. Multiple files are separated by comma.",
            required=True,
        )
        parser.add_argument(
            "--fq2_3p",
            help="3 prime R2 fastq file. Multiple files are separated by comma.",
            required=True,
        )
        parser.add_argument(
            "--fq1_5p",
            help="5 prime R1 fastq file. Multiple files are separated by comma.",
            required=True,
        )
        parser.add_argument(
            "--fq2_5p",
            help="5 prime R2 fastq file. Multiple files are separated by comma.",
            required=True,
        )
        parser = s_common(parser)

    return parser



import collections
import math

import numpy as np
import pandas as pd
import plotly.graph_objs as go
import plotly.offline as pltoff

from celescope.tools.utils import add_log

BarcodeRankPlotSegment = collections.namedtuple(
    "BarcodeRankPlotSegment", ["start", "end", "cell_density", "legend"]
)

CHARTS_PLOTLY_MODEBAR_TRANSFORM_BUTTONS = [
    "zoom2d",
    "pan2d",
    "zoomIn2d",
    "zoomOut2d",
    "autoScale2d",
    # 'resetScale2d'  can't totally disable interaction, it seems-- keep reset option
]

CHARTS_PLOTLY_EXPORT_BUTTONS = [
    "toImage",
    "sendDataToCloud",
]

CHARTS_PLOTLY_FIXED_CONFIG = {
    "modeBarButtonsToRemove": CHARTS_PLOTLY_MODEBAR_TRANSFORM_BUTTONS
    + CHARTS_PLOTLY_EXPORT_BUTTONS,
    "displaylogo": False,
    "showLink": False,
}

CHARTS_PLOTLY_MOVABLE_CONFIG = {
    "modeBarButtonsToRemove": CHARTS_PLOTLY_EXPORT_BUTTONS,
    "displaylogo": False,
    "showLink": False,
}
BC_RANK_PLOT_LINE_WIDTH = 3
# Gradient scheme used in the barcode rank plot
BC_PLOT_COLORS = [
    "#dddddd",
    "#d1d8dc",
    "#c6d3dc",
    "#bacfdb",
    "#aecada",
    "#a3c5d9",
    "#97c0d9",
    "#8cbbd8",
    "#80b7d7",
    "#74b2d7",
    "#6aadd6",
    "#66abd4",
    "#62a8d2",
    "#5ea5d1",
    "#59a2cf",
    "#559fce",
    "#519ccc",
    "#4d99ca",
    "#4997c9",
    "#4594c7",
    "#4191c5",
    "#3d8dc4",
    "#3a8ac2",
    "#3787c0",
    "#3383be",
    "#3080bd",
    "#2c7cbb",
    "#2979b9",
    "#2676b7",
    "#2272b6",
    "#1f6eb3",
    "#1d6ab0",
    "#1a65ac",
    "#1861a9",
    "#155ca6",
    "#1358a2",
    "#10539f",
    "#0e4f9b",
    "#0b4a98",
    "#094695",
    "#09438f",
    "#0a4189",
    "#0c3f83",
    "#0d3d7c",
    "#0e3b76",
    "#103970",
    "#11366a",
    "#123463",
    "#14325d",
    "#153057",
]

CHARTS = [
    {
        "layout": {
            "title": "Barcode Rank",
            "width": 470,
            "height": 313,
            "margin": {"l": 60, "r": 0, "t": 30, "b": 40},
            "hovermode": "closest",
            "xaxis": {
                "title": "Barcodes",
                "type": "log",
            },
            "yaxis": {
                "title": "UMI counts",
                "type": "log",
            },
        },
        "data": [],
        "config": CHARTS_PLOTLY_FIXED_CONFIG,
        "function": "_plot_barcode_rank",
        "description": "The plot shows the count of filtered UMIs mapped to each barcode.  Barcodes can be determined to be cell-associated based on their UMI counts or by their expression profiles. Therefore some regions of the graph can contain both cell-associated and background-associated barcodes. The color of the graph represents the local density of barcodes that are cell-associated.",
        "name": "barcode_rank",
    },
]


def BC_PLOT_CMAP(density):
    """
    Colormap utility fn to map a number to one of the colors in the gradient
    color scheme defined above
    Input
    - density : A real number in the range [0,1]
    """
    assert density >= 0.0
    assert density <= 1.0
    levels = len(BC_PLOT_COLORS)
    ind = min(levels - 1, int(math.floor(levels * density)))
    return BC_PLOT_COLORS[ind]


def get_plot_segment(start_index, end_index, sorted_bc, cell_barcodes, legend=False):
    """
    Helper function to build a plot segment.
    """
    assert end_index > start_index
    num_cells = sum(
        [1 for i in range(start_index, end_index) if sorted_bc[i] in cell_barcodes]
    )
    density = float(num_cells) / float(end_index - start_index)
    return BarcodeRankPlotSegment(
        start=start_index, end=end_index, cell_density=density, legend=legend
    )


def segment_log_plot_by_length(y_data, x_start, x_end):
    """
    Given the extends of the mixed region [x_start, x_end), compute
    the x-indices that would divide the plot into segments of a
    prescribed length (in pixel coordinates) with a minimum number
    of barcodes in each segment
    """

    if x_end <= x_start:
        return []

    SEGMENT_NORMALIZED_MAX_LEN = 0.02
    MIN_X_SPAN = 20

    log_max_x = np.log(len(y_data))
    log_max_y = np.log(max(y_data))

    this_segment_len = 0.0
    segment_idx = [x_start]

    for i in range(x_start, x_end):
        last_i = max(x_start, i - 1)
        dx = (np.log(i) - np.log(last_i)) / log_max_x
        dy = (np.log(y_data[i]) - np.log(y_data[last_i])) / log_max_y
        this_segment_len += np.linalg.norm([dx, dy])
        if this_segment_len >= SEGMENT_NORMALIZED_MAX_LEN and i > (
            segment_idx[-1] + MIN_X_SPAN
        ):
            segment_idx.append(i + 1)
            this_segment_len = 0.0

    if segment_idx[-1] != x_end:
        segment_idx.append(x_end)

    return segment_idx


def segment_log_plot_by_length_log_uniform(x_start, x_end):
    """
    vdj usually have very large x_end
    """

    if x_end <= x_start:
        return []

    MIN_X_LOG10_SPAN = 0.03
    MIN_BARCODE = 20
    MAX_N = 100
    segment_idx = [x_start]
    span = np.log10(x_end - x_start) / MAX_N
    if span < MIN_X_LOG10_SPAN:
        span = MIN_X_LOG10_SPAN

    span_sum = 1
    idx = x_start
    while True:
        span_sum += span
        idx = int(x_start + 10**span_sum)
        if idx - segment_idx[-1] < MIN_BARCODE:
            continue
        if idx > x_end:
            break
        segment_idx.append(idx)

    if segment_idx[-1] != x_end:
        segment_idx.append(x_end)

    return segment_idx


def convert_numpy_array_to_line_chart(array, ntype):
    array = np.sort(array)[::-1]

    rows = []
    previous_count = None
    for (index,), count in np.ndenumerate(array):
        if index == 0 or index == len(array) - 1:
            rows.append([index, ntype(count)])
        elif previous_count != count:
            previous_index = rows[-1][0]
            if previous_index != index - 1:
                rows.append([index - 1, ntype(previous_count)])
            rows.append([index, ntype(count)])
        previous_count = count
    return rows


@add_log
def counter_barcode_rank_plot_data(df_UMI_path, log_uniform=False):
    """
    get cell density for each plot_segments
    :param count_data_path:
    :return: sorted_counts, plot_segments, cell_nums
    """
    df_UMI = pd.read_csv(df_UMI_path, index_col=0, sep="\t")
    cell_bc = set(df_UMI[df_UMI["mark"] == "CB"].index)
    sorted_bc = np.array(df_UMI.index)
    sorted_counts = np.array(df_UMI["UMI"])
    cell_nums = len(cell_bc)
    total_bc = len(sorted_bc)
    # find the first barcode which is not a cell
    first_non_cell = total_bc
    for i, bc in enumerate(sorted_bc):
        if bc not in cell_bc:
            first_non_cell = i
            break

    # find the last barcode which is a cell
    last_cell = 0
    for i in reversed(range(total_bc)):
        if sorted_bc[i] in cell_bc:
            last_cell = i
            break

    ranges = [0, first_non_cell, last_cell + 1, total_bc]
    plot_segments = []
    plot_segments.append(
        BarcodeRankPlotSegment(start=0, end=ranges[1], cell_density=1.0, legend=True)
    )
    plot_segments.append(
        BarcodeRankPlotSegment(
            start=ranges[2], end=ranges[3], cell_density=0.0, legend=True
        )
    )

    if not log_uniform:
        mixed_segments = segment_log_plot_by_length(sorted_counts, ranges[1], ranges[2])
    else:
        mixed_segments = segment_log_plot_by_length_log_uniform(ranges[1], ranges[2])
    for i in range(len(mixed_segments) - 1):
        plot_segments.append(
            get_plot_segment(
                mixed_segments[i],
                mixed_segments[i + 1],
                sorted_bc,
                cell_bc,
                legend=False,
            )
        )

    return sorted_counts, plot_segments, cell_nums


def _plot_barcode_rank(chart, counts, num_cells):
    """Generate a generic barcode rank plot"""
    rows = convert_numpy_array_to_line_chart(counts, int)

    for _i, row in enumerate(rows):
        index, count = row[0], row[1]
        if index < num_cells:
            series_list = [chart["data"][0]]
        elif index == num_cells:
            # Connect the two lines
            series_list = [chart["data"][0], chart["data"][1]]
        else:
            series_list = [chart["data"][1]]

        for series in series_list:
            series["x"].append(index)
            series["y"].append(count)

    # Handle case where there is no background
    bg_series = chart["data"][1]
    if len(bg_series["x"]) == 0:
        bg_series["x"].append(0)
        bg_series["y"].append(0)

    return chart


def build_plot_data_dict(plot_segment, counts):
    """
    Construct the data for a plot segment by appropriately slicing the
    counts
    Inputs:
    - plot_segment: BarcodeRankPlotSegment containing [start, end)
        of the segment, the cell density and legend visibility option
    - counts: Reverse sorted UMI counts for all barcodes.
    """

    start = max(0, plot_segment.start - 1)  # -1 for continuity between two charts
    end = plot_segment.end
    plot_rows = convert_numpy_array_to_line_chart(counts[start:end], int)
    name = "Cells" if plot_segment.cell_density > 0 else "Background"

    # Setup the tooltip
    if plot_segment.cell_density > 0.0:
        n_barcodes = plot_segment.end - plot_segment.start
        n_cells = int(round(plot_segment.cell_density * n_barcodes))
        hover = "{:.0f}% Cells<br>({}/{})".format(
            100 * plot_segment.cell_density, n_cells, n_barcodes
        )
    else:
        hover = "Background"

    data_dict = {
        "x": [],
        "y": [],
        "name": name,
        "hoverinfo": "text",
        "text": hover,
        "type": "scattergl",
        "mode": "lines",
        "line": {
            "width": 3,
            "color": BC_PLOT_CMAP(plot_segment.cell_density),
        },
        "showlegend": plot_segment.legend,
    }
    offset = 1 + start  # it's a log-log plot, hence the 1
    for index, count in plot_rows:
        data_dict["x"].append(index + offset)
        data_dict["y"].append(count)

    # Handle case where the data is empty
    if len(data_dict["x"]) == 0:
        data_dict["x"].append(0)
        data_dict["y"].append(0)

    return data_dict


def _plot_counter_barcode_rank(chart, counts, plot_segments):
    """
    Generate the RNA counter barcode rank plot
    Inputs:
        - chart: chart element to populate data
        - counts: UMI counts reverse sorted
        - plot_segments: A list of BarcodeRankPlotSegments
    """

    for segment in plot_segments:
        chart["data"].append(build_plot_data_dict(segment, counts))

    return chart


@add_log
def get_plot_data(plot_segments, counts):
    plot_data = []
    for segment in plot_segments:
        plot_data.append(build_plot_data_dict(segment, counts))

    return plot_data


def plot_barcode_rank(count_file_path, log_uniform=False):
    sorted_counts, plot_segments, _cell_nums = counter_barcode_rank_plot_data(
        count_file_path, log_uniform=log_uniform
    )
    plot_data = get_plot_data(plot_segments, sorted_counts)

    plotly_data = [
        go.Scatter(
            x=dat["x"],
            y=dat["y"],
            name=dat["name"],
            mode=dat["mode"],
            showlegend=dat["showlegend"],
            marker={"color": dat["line"]["color"]},
            line=dat["line"],
            text=dat["text"],
        )
        for dat in plot_data
    ]

    layout = go.Layout(
        width=470,
        height=313,
        title={"text": "Barcode rank", "font": {"color": "black"}, "x": 0.5},
        xaxis={
            "type": "log",
            "title": "Barcode",
            "titlefont": {"color": "black"},
            "color": "black",
            "gridcolor": "gainsboro",
            "linecolor": "black",
        },
        yaxis={
            "type": "log",
            "title": "UMI counts",
            "titlefont": {"color": "black"},
            "color": "black",
            "gridcolor": "gainsboro",
            "linecolor": "black",
        },
        margin=dict(l=50, r=0, t=30, b=30),
        plot_bgcolor="#FFFFFF",
    )

    config = dict(
        {
            "displayModeBar": True,
            "staticPlot": False,
            "showAxisDragHandles": False,
            "modeBarButtons": [["toImage", "resetScale2d"]],
            "scrollZoom": False,
            "displaylogo": False,
        }
    )

    fig = go.Figure(data=plotly_data, layout=layout)

    chart = pltoff.plot(fig, include_plotlyjs=False, output_type="div", config=config)

    return chart



STEPS = ["sample", "starsolo", "analysis", "conversion", "substitution", "replacement"]

__ASSAY__ = "dynaseq"

IMPORT_DICT = {
    "analysis": "celescope.rna",
}


# count
DYNA_MATRIX_DIR_SUFFIX = ["labeled", "unlabeled"]



# count
OUTS_DIR = "outs"
RAW_MATRIX_DIR_SUFFIX = "raw"
FILTERED_MATRIX_DIR_SUFFIX = "filtered"
MATRIX_FILE_NAME = "matrix.mtx.gz"
FEATURE_FILE_NAME = "features.tsv.gz"
BARCODE_FILE_NAME = "barcodes.tsv.gz"
STAR_BAM_SUFFIX = "Aligned.out.bam"
TAG_BAM_SUFFIX = "aligned_posSorted_addTag.bam"
STARSOLO_BAM_SUFFIX = "Aligned.sortedByCoord.out.bam"
COUNTS_FILE_NAME = "counts.tsv"

# mkref
GENOME_CONFIG = "celescope_genome.config"



from celescope.tools import utils
from celescope.__init__ import HELP_DICT
from celescope.tools.step import Step, s_common
from celescope.rna.mkref import Mkref_rna
from celescope.snp.__init__ import PANEL


class Variant_calling(Step):
    """
    ## Features
    - Perform variant calling at single cell level.

    ## Output
    - `{sample}_raw.vcf` Variants are called with bcftools default settings.
    - `{sample}_norm.vcf` Indels are left-aligned and normalized. See https://samtools.github.io/bcftools/bcftools.html#norm for more details.
    """

    def __init__(self, args):
        Step.__init__(self, args)

        # set
        self.barcodes, _num = utils.get_barcode_from_match_dir(args.match_dir)
        self.fasta = Mkref_rna.get_config(args.genomeDir)["files"]["fasta"]
        self.df_vcf = None
        self.panel = args.panel
        self.bed = utils.get_bed_file_path(self.panel)

        # out
        self.splitN_bam = f"{self.out_prefix}_splitN.bam"
        self.splitN_bam_name_sorted = f"{self.out_prefix}_splitN_name_sorted.bam"

        self.raw_bcf_file = f"{self.out_prefix}_raw.bcf"
        self.raw_vcf_file = f"{self.out_prefix}_raw.vcf"
        self.fixed_header_vcf = f"{self.out_prefix}_fixed.vcf"
        self.norm_vcf_file = f"{self.out_prefix}_norm.vcf"

    @utils.add_log
    def SplitNCigarReads(self):
        cmd = (
            f"gatk "
            f"SplitNCigarReads "
            f"--do-not-fix-overhangs "
            f"-R {self.fasta} "
            f"-I {self.args.bam} "
            f"-O {self.splitN_bam} "
        )
        self.debug_subprocess_call(cmd)

    @utils.add_log
    def gatk_norm(self):
        cmd = (
            f"gatk LeftAlignAndTrimVariants "
            f"-R {self.fasta} "
            f"-V {self.fixed_header_vcf} "
            f"-O {self.norm_vcf_file} "
            "--split-multi-allelics "
        )
        self.debug_subprocess_call(cmd)

    @utils.add_log
    def call_variants(self):
        """
        https://github.com/samtools/bcftools/issues/1459
        """
        cmd = (
            f"bcftools mpileup "
            f"-f {self.fasta} "
            f"--threads {self.thread} "
            f"--annotate DP,AD "
            f"--max-depth 2000000 --max-idepth 2000000 --no-BAQ "
            f"-o {self.raw_bcf_file} "
            f"{self.splitN_bam} "
        )
        if self.bed:
            cmd += f" --regions-file {self.bed} "
        self.debug_subprocess_call(cmd)

        cmd = (
            f"bcftools call "
            f"-mv -Ov "
            f"-o {self.raw_vcf_file} "
            f"{self.raw_bcf_file} "
        )
        self.debug_subprocess_call(cmd)

    def bcftools_norm(self):
        cmd = (
            "bcftools norm "
            "-m- "
            f"-f {self.fasta} "
            f"{self.raw_vcf_file} "
            "| bcftools norm "
            "-d both "
            f"-o {self.norm_vcf_file} "
        )
        self.debug_subprocess_call(cmd)

    def run(self):
        self.SplitNCigarReads()
        self.call_variants()
        self.bcftools_norm()


@utils.add_log
def variant_calling(args):
    with Variant_calling(args) as runner:
        runner.run()


def get_opts_variant_calling(parser, sub_program):
    parser.add_argument("--genomeDir", help=HELP_DICT["genomeDir"], required=True)
    parser.add_argument("--panel", help=HELP_DICT["panel"], choices=list(PANEL))
    if sub_program:
        parser.add_argument(
            "--bam", help="Input BAM file from step `target_metrics`. ", required=True
        )
        parser.add_argument("--match_dir", help=HELP_DICT["match_dir"], required=True)
        s_common(parser)



import subprocess
import sys

import numpy as np
import pandas as pd

from celescope.tools.__init__ import COUNTS_FILE_NAME
from celescope.tools.emptydrop_cr import get_plot_elements
from celescope.tools.starsolo import (
    Demultiplexing,
    Mapping as ToolsMapping,
    create_solo_args,
    create_soloFeatures,
    get_opts_starsolo as tools_opts,
)
from celescope.tools.starsolo import (
    Starsolo as tools_Starsolo,
)
from celescope.tools.matrix import CountMatrix
from celescope.tools import utils
from celescope.tools.step import Step

SAM_attributes = "NH HI nM AS CR UR CB UB GX GN "


def get_well_barcode(
    bc_file: str,
) -> dict[int, str]:
    barcodes = utils.one_col_to_list(bc_file)
    return {i: x for i, x in enumerate(barcodes, start=1)}


def get_barcode_sample(
    bc_file: str,
    well_sample_file: str,
) -> dict[str, str]:
    well_barcode = get_well_barcode(bc_file)
    well_sample = utils.two_col_to_dict(well_sample_file)
    barcode_sample = {
        well_barcode[well]: sample for well, sample in well_sample.items()
    }
    return barcode_sample


class Starsolo(tools_Starsolo):
    def __init__(self, args, display_title=None):
        super().__init__(args, display_title=display_title)
        self.well_barcode = get_well_barcode(self.bc[0])
        self.barcode_sample = get_barcode_sample(self.bc[0], args.well_sample)

        self.tsv_matrix_file = f"{self.out_prefix}_matrix.tsv.gz"
        self.outs.append(self.tsv_matrix_file)

    def run_starsolo(self):
        if "--outSAMunmapped Within" not in self.extra_starsolo_args:
            self.extra_starsolo_args += " --outSAMunmapped Within "
        soloFeatures = create_soloFeatures(
            self.args.soloFeatures, self.args.report_soloFeature
        )
        cmd = create_solo_args(
            pattern_args=self.pattern_args,
            whitelist_args=self.whitelist_args,
            outFileNamePrefix=self.out_prefix + "_",
            fq1=self.args.fq1,
            fq2=self.args.fq2,
            genomeDir=self.args.genomeDir,
            soloCellFilter="None",  # only use raw matrix
            runThreadN=self.args.thread,
            clip3pAdapterSeq=self.args.adapter_3p,
            outFilterMatchNmin=self.args.outFilterMatchNmin,
            soloFeatures=soloFeatures,
            outSAMtype=self.args.outSAMtype,
            outSAMattributes=self.outSAMattributes,
            soloCBmatchWLtype=self.args.soloCBmatchWLtype,
            limitBAMsortRAM=self.args.limitBAMsortRAM,
            extra_starsolo_args=self.extra_starsolo_args,
        )
        if self.chemistry == "bulk_rna-bulk_vdj_match":
            cmd += "--soloStrand Reverse \\\n"
        sys.stderr.write(cmd)
        subprocess.check_call(cmd, shell=True)
        cmd = f"chmod -R 755 {self.solo_out_dir}"
        sys.stderr.write(cmd)
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def keep_barcodes(self):
        """take in raw matrix, only keep barcodes in the input file, and convert barcodes to sample names"""
        matrix = CountMatrix.from_matrix_dir(self.raw_matrix)

        filtered = matrix.slice_matrix_bc(self.barcode_sample.keys())
        filtered.to_matrix_dir(self.filtered_matrix)
        samples = [self.barcode_sample[bc] for bc in filtered.get_barcodes()]
        converted = CountMatrix(filtered.get_features(), samples, filtered.get_matrix())
        df = converted.to_df()
        df.to_csv(self.tsv_matrix_file, sep="\t")
        return filtered

    def run(self):
        self.run_starsolo()
        filtered = self.keep_barcodes()
        self.gzip_matrix()
        q30_cb, q30_umi = self.get_Q30_cb_UMI()
        return q30_cb, q30_umi, filtered, self.barcode_sample, self.well_barcode


class Mapping(ToolsMapping):
    # only add metrics
    def __init__(self, args, display_title=None):
        super().__init__(args, display_title=display_title)

    @utils.add_log
    def parse_cellReadsStats(self) -> tuple[dict[str, int], pd.DataFrame]:
        df = pd.read_csv(self.cellReadsStats, sep="\t", header=0, index_col=0)
        df = df.iloc[1:,]  # skip first line cb not pass whitelist
        df_count = df.loc[:, ["nUMIunique", "countedU"]]  # keep dataframe format
        df_count.rename(columns={"nUMIunique": "UMI"}, inplace=True)
        df_count.sort_values(by="UMI", ascending=False, inplace=True)
        cbs = CountMatrix.read_barcodes(self.filtered_matrix)
        df_count["mark"] = "UB"
        for cb in cbs:
            df_count.loc[cb, "mark"] = "CB"
        df_count.to_csv(self.counts_file, sep="\t", index=True)

        df = df.loc[
            :,
            [
                "cbMatch",
                "cbPerfect",
                "genomeU",
                "genomeM",
                "exonic",
                "intronic",
                "exonicAS",
                "intronicAS",
                "countedU",
            ],
        ]
        metrics = df.sum().to_dict()
        df = df.loc[
            :,
            [
                "cbMatch",
                "genomeU",
                "genomeM",
            ],
        ]
        df.rename(
            columns={
                "cbMatch": "Reads",
                "genomeU": "Unique-mapping Reads",
                "genomeM": "Multiple-mapping Reads",
            },
            inplace=True,
        )
        df["Unique-mapping Reads"] = np.where(
            df["Reads"] != 0,
            (df["Unique-mapping Reads"] / df["Reads"] * 100).round(2).astype(str) + "%",
            np.nan,
        )
        df["Multiple-mapping Reads"] = np.where(
            df["Reads"] != 0,
            (df["Multiple-mapping Reads"] / df["Reads"] * 100).round(2).astype(str)
            + "%",
            np.nan,
        )
        return metrics, df

    @utils.add_log
    def run(self):
        metrics, df = self.parse_cellReadsStats()
        return self.add_metrics_to_report(metrics), df


class Cells(Step):
    def __init__(self, args, display_title=None):
        super().__init__(args, display_title=display_title)
        solo_dir = (
            f"{self.outdir}/{self.sample}_Solo.out/{self.args.report_soloFeature}"
        )
        self.summary_file = f"{solo_dir}/Summary.csv"
        self.counts_file = f"{self.outs_dir}/{COUNTS_FILE_NAME}"

    @utils.add_log
    def parse_summary(self):
        df = pd.read_csv(self.summary_file, index_col=0, header=None)
        s = df.iloc[:, 0]
        saturation = float(s["Sequencing Saturation"])
        n_reads = int(s["Number of Reads"])
        q30_RNA = float(s["Q30 Bases in RNA read"])

        return n_reads, q30_RNA, saturation

    def run(
        self,
        filtered: CountMatrix,
        barcode_sample: dict,
        well_barcode: dict,
        df_metrics: pd.DataFrame,
    ):
        df_counts = pd.read_csv(self.counts_file, index_col=0, header=0, sep="\t")
        reads_total = df_counts["countedU"].sum()
        bcs = filtered.get_barcodes()
        n_cells = len(bcs)
        reads_cell = df_counts.loc[bcs, "countedU"].sum()
        fraction_reads_in_cells = float(reads_cell / reads_total)
        mean_used_reads_per_cell = int(reads_cell // len(bcs))
        median_umi_per_cell = int(df_counts.loc[bcs, "UMI"].median())

        bc_geneNum, total_genes = filtered.get_bc_geneNum()
        median_genes_per_cell = int(np.median(list(bc_geneNum.values())))

        df_counts.loc[:, "mark"] = "UB"
        df_counts.loc[bcs, "mark"] = "CB"
        df_counts.fillna(0, inplace=True)
        df_counts = df_counts.astype({"UMI": int, "countedU": int})
        df_counts.to_csv(self.counts_file, sep="\t", index=True)

        self.add_metric(
            "Number of Wells",
            n_cells,
            help_info="number of barcodes with at least one UMI",
        )
        self.add_metric(
            "Fraction of Reads in Wells",
            fraction_reads_in_cells,
            value_type="fraction",
            help_info="Fraction of reads which were mapped to a barcode",
        )
        self.add_metric(
            "Mean Used Reads per Well",
            mean_used_reads_per_cell,
            help_info="The number of uniquely-mapped-to-transcriptome reads per well",
        )
        self.add_metric(
            "Median UMI per Well",
            median_umi_per_cell,
            help_info="Median UMI count per barcode",
        )
        self.add_metric(
            "Median Genes per Well",
            median_genes_per_cell,
            help_info="Median number of genes per barcode",
        )
        self.add_data(chart=get_plot_elements.plot_barcode_rank(self.counts_file))

        n_reads, q30_RNA, saturation = self.parse_summary()
        self.add_metric(
            "Saturation",
            saturation,
            value_type="fraction",
            help_info="the fraction of read originating from an already-observed UMI.",
        )

        # table
        df_cells = df_counts[df_counts["mark"] == "CB"]
        df_cells["Sample"] = df_cells.index.map(lambda x: barcode_sample[x])
        df_cells["Genes"] = df_cells.index.map(lambda x: bc_geneNum.get(x, 0))
        df_cells["Barcode"] = df_cells.index
        barcode_well = {v: k for k, v in well_barcode.items()}
        df_cells["Well"] = df_cells["Barcode"].map(lambda x: barcode_well[x])
        df_cells = df_cells.merge(
            df_metrics, left_on="Barcode", right_index=True, how="left"
        )
        df_cells = df_cells.loc[
            :,
            [
                "Well",
                "Sample",
                "Barcode",
                "Reads",
                "Unique-mapping Reads",
                "Multiple-mapping Reads",
                "UMI",
                "Genes",
            ],
        ]

        table_dict = self.get_table_dict(
            title="Metrics for a Single Well",
            table_id=self.assay,
            df_table=df_cells,
        )
        self.add_data(table_dict=table_dict)

        return n_reads, q30_RNA


def starsolo(args):
    with Starsolo(args) as runner:
        q30_cb, q30_umi, filtered, barcode_sample, well_barcode = runner.run()

    with Mapping(args) as runner:
        (valid_reads, corrected), df_metrics = runner.run()

    with Cells(args, display_title="Wells") as runner:
        n_reads, q30_RNA = runner.run(
            filtered, barcode_sample, well_barcode, df_metrics
        )

    with Demultiplexing(args) as runner:
        runner.run(valid_reads, n_reads, corrected, q30_cb, q30_umi, q30_RNA)


def get_opts_starsolo(parser, sub_program=True):
    tools_opts(parser, sub_program)
    parser.add_argument(
        "--well_sample",
        help="tsv file of well numbers and sample names. The first column is well numbers and the second column is sample names.",
        required=True,
    )
    return parser



import pandas as pd

from celescope.tools.step import Step, s_common
from celescope.tools.plotly_plot import Tsne_dropdown_plot, Tsne_plot


def get_opts_analysis_cite(parser, sub_program):
    if sub_program:
        parser.add_argument("--tsne_coord", help="tsne coord file", required=True)
        s_common(parser)


def analysis_cite(args):
    with Analysis_cite(args, display_title="Analysis") as runner:
        runner.run()


class Analysis_cite(Step):
    def __init__(self, args, display_title):
        super().__init__(args, display_title)

    def run(self):
        df_tsne = pd.read_csv(self.args.tsne_coord, sep="\t", index_col=0)
        # filter low expression
        keep_antibody = df_tsne.columns[4:][
            (df_tsne.iloc[:, 4:].mean() > 0.1).to_list()
        ].to_list()
        keep_cols = df_tsne.columns[:4].to_list() + keep_antibody
        df_tsne = df_tsne.loc[:, keep_cols]

        # accelerate
        df_tsne["tSNE_1"] = df_tsne["tSNE_1"].astype("float16")
        df_tsne["tSNE_2"] = df_tsne["tSNE_2"].astype("float16")
        for col in keep_antibody:
            df_tsne[col] = df_tsne[col].astype("float16")

        # plot
        tsne_cluster = Tsne_plot(df_tsne, "cluster").get_plotly_div()
        self.add_data(tsne_cluster=tsne_cluster)
        tsne_citeseq = Tsne_dropdown_plot(
            df_tsne, "Citeseq", keep_antibody
        ).get_plotly_div()
        self.add_data(tsne_citeseq=tsne_citeseq)



from celescope.tools.tag.count_tag import (
    Count_tag as Ct,
    count_tag as ct,
    get_opts_count_tag as opts,
)


class Count_tag(Ct):
    pass


def count_tag(args):
    ct(args)


def get_opts_count_tag(parser, sub_program):
    opts(parser, sub_program)



import argparse
import glob
import os


class Debug:
    def __init__(self):
        self.outdir = "debug"
        if not os.path.exists(self.outdir):
            os.mkdir(self.outdir)
        self.opts()
        self.all_reads = glob.glob(f"{self.dir}/02.cutadapt/*clean_2.fq.gz")[0]
        # out files
        self.sub_reads = f"{self.outdir}/sub.fq.gz"
        self.unmapped = f"{self.outdir}/sub_Unmapped.out.mate1"

    def opts(self):
        readme = "debug"
        parser = argparse.ArgumentParser(readme)
        parser.add_argument("--dir", required=True)
        parser.add_argument("--n_sub", default=1000000)
        parser.add_argument(
            "--genomeDir",
            default="/SGRNJ/Public/Database/genome/homo_sapiens/ensembl_92",
        )
        args = parser.parse_args()
        self.dir = args.dir
        self.n_sub = args.n_sub
        self.genomeDir = args.genomeDir
        self.thread = 2

    def run_subsample(self):
        cmd = (
            f"less {self.all_reads} "
            f"|head -n {self.n_sub * 4} | gzip -c > {self.sub_reads} "
        )
        with open("subsample.sh", "wt") as f:
            f.write(cmd)

    def run_STAR(self):
        cmd = (
            f"star --runThreadN {self.thread} \\\n"
            f"--genomeDir {self.genomeDir} \\\n"
            f"--readFilesIn {self.sub_reads} \\\n"
            f"--readFilesCommand zcat --outFilterMultimapNmax 1 \\\n"
            f"--outFileNamePrefix {self.outdir}/sub_ \\\n"
            f"--outSAMtype BAM SortedByCoordinate --outFilterMatchNmin 0 \\\n"
            f"--outReadsUnmapped Fastx \\\n"
        )
        with open("STAR.sh", "wt") as f:
            f.write(cmd)

    def fastqc(self):
        cmd = f"fastqc {self.unmapped} \\\n"
        with open("fastqc.sh", "wt") as f:
            f.write(cmd)

    def run(self):
        self.run_subsample()
        self.run_STAR()
        self.fastqc()


if __name__ == "__main__":
    de = Debug()
    de.run()
